
--[[

Puzzlet

actions.lua

(c) bob fossil 2016

--]]

require "audio"
require "collisions"
require "weapons"

MOVEMENT = 2
EXIT_ANIMATION_TIME = 500
KEY_REPEAT_TIME_INTERVAL = 150



--

function can_move_into_tile(map_x, map_y)

	-- Returns true if we can move the player into the specified tile.
	local value = global.map[map_y][map_x]
	
	-- Bail out if the tile is in the walls or door range.
	if value > LEVEL_EXIT_RANGE then
		if value > LEVEL_SPECIAL_RANGE then
			if value < LEVEL_PLAYER_RANGE then
				-- Tile is a special tile.
				if value ~= GRAPHIC_SPECIAL_BONES or
					value ~= GRAPHIC_SPECIAL_GENERATOR_DAMAGED or
					value ~= GRAPHIC_SPECIAL_GENERATOR_OFF or
					value ~= GRAPHIC_SPECIAL_GENERATOR_ON then
						return true
				end
				
				if value == GRAPHIC_SPECIAL_GROUND_FORCEFIELD then
					if global.current_player==PLAYER_VALKYRIE then
						return true
					end
				end
			end
		else 
			-- Tile is an exit tile.
			if value < GRAPHIC_EXIT_APPEAR_1 then
				return true
			end
		end
	end

	return false
end

function can_teleport(area)

	local tx = (global.areas[area].left / TILE_WIDTH) + 1
	local ty = (global.areas[area].top / TILE_HEIGHT) + 1

	-- See if tile is still a destination point.
	if global.map[ty][tx]< GRAPHIC_SPECIAL_AREA_1 + LEVEL_SPECIAL_RANGE or
		global.map[ty][tx] > GRAPHIC_SPECIAL_AREA_8 + LEVEL_SPECIAL_RANGE then
		dprintf("Can't teleport. No teleport destination found! %d", global.map[ty][tx])
		return false
	end
	
	-- Check another player isn't in the area we're
	-- about to teleport into.
	for check = 1, MAX_PLAYERS do

		if check~=global.current_player and
			global.players[check].active then
	--[[
			dprintf("Checking player %d", check)

			dprintf("%d %d %d %d - %d %d %d %d", global.players[check].left, global.players[check].top, global.players[check].right, global.players[check].bottom,
				global.areas[area].left, global.areas[area].top, global.areas[area].right, global.areas[area].bottom)
	--]]									
			if (rect_intersect_rect(global.players[check], global.areas[area]) or
				rect_equal_rect(global.players[check], global.areas[area])) then
			
				dprintf("Can't teleport. Player in area!")
				return false
			end
		end
	end
	
	for check = 1, #global.enemies do

		if (rect_intersect_rect(global.enemies[check], global.areas[area]) or
			rect_equal_rect(global.enemies[check], global.areas[area])) then
		
			dprintf("Can't teleport. Enemy in area!")
			return false
		end
	end
	
	return true
end

function handle_wrap_around(rect, direction)

	local wrapped = false
	local wrapped_rect = {left = rect.left, top = rect.top, right = rect.right, bottom = rect.bottom}
	
	-- Handle wrapping.		
	if rect.left < 0 or rect.top < 0 or
		rect.right > MAX_TILE_WIDTH or rect.bottom > MAX_TILE_HEIGHT then

		local restore_rect = global.players[global.current_player]	
		local wrap_direction = direction
		--dprintf("Rectangle wrapped")

		-- Strip out diagonal movements. We can't wrap on diagonals.
		if wrap_direction==UP_LEFT then
			if rect.left < 0 then
				wrap_direction = LEFT
				wrapped_rect.top = restore_rect.top
				wrapped_rect.bottom = restore_rect.bottom
			elseif rect.top < 0 then
				wrap_direction = UP
				wrapped_rect.left = restore_rect.left
				wrapped_rect.right = restore_rect.right
			end
		elseif wrap_direction==UP_RIGHT then
			if rect.right > MAX_TILE_WIDTH then
				wrap_direction = RIGHT
				wrapped_rect.top = restore_rect.top
				wrapped_rect.bottom = restore_rect.bottom
			elseif rect.top < 0 then
				wrap_direction = UP
				wrapped_rect.left = restore_rect.left
				wrapped_rect.right = restore_rect.right				
			end		
		elseif wrap_direction==DOWN_LEFT then
			if rect.left < 0 then
				wrap_direction = LEFT
				wrapped_rect.top = restore_rect.top
				wrapped_rect.bottom = restore_rect.bottom
			elseif rect.bottom > MAX_TILE_HEIGHT then
				wrap_direction = DOWN
				wrapped_rect.left = restore_rect.left
				wrapped_rect.right = restore_rect.right				
			end		
		elseif wrap_direction==DOWN_RIGHT then
			if rect.right > MAX_TILE_WIDTH then
				wrap_direction = RIGHT
				wrapped_rect.top = restore_rect.top
				wrapped_rect.bottom = restore_rect.bottom
			elseif rect.bottom > MAX_TILE_HEIGHT then
				wrap_direction = DOWN
				wrapped_rect.left = restore_rect.left
				wrapped_rect.right = restore_rect.right				
			end		
		end
		
		if wrap_direction==UP then

			--dprintf("Rectangle wrapped top")
			wrapped_rect.top = MAX_TILE_HEIGHT - TILE_HEIGHT
			wrapped_rect.bottom = MAX_TILE_HEIGHT - 1
			wrapped = true
	
		elseif wrap_direction==DOWN then

			--dprintf("Rectangle wrapped down")
			wrapped_rect.top = 0
			wrapped_rect.bottom = TILE_HEIGHT - 1
			wrapped = true
			
		elseif wrap_direction==LEFT then

			--dprintf("Rectangle wrapped left")
			wrapped_rect.left = MAX_TILE_WIDTH - TILE_WIDTH
			wrapped_rect.right = MAX_TILE_WIDTH - 1
			
			--dprintf("%d %d", wrapped_rect.top, wrapped_rect.bottom)
			
			wrapped = true

		elseif wrap_direction==RIGHT then

			--dprintf("Rectangle wrapped right")
			wrapped_rect.left = 0
			wrapped_rect.right = TILE_WIDTH - 1
			wrapped = true			
		end		
	end
	
	return wrapped, wrapped_rect
end

function handle_player_wrapped(player_rect, direction)

	local wrapped, wrapped_rect = handle_wrap_around(player_rect, direction)
	
	if wrapped then
		-- We've wrapped. But we only appear on the other side if we can move into
		-- the tile.
		local wrap_x = math.floor(wrapped_rect.left / TILE_WIDTH) + 1
		local wrap_y = math.floor(wrapped_rect.top / TILE_HEIGHT) + 1
		
		local wrap_collision = check_for_tile_collision(wrap_x, wrap_y)
		if wrap_collision==COLLISION_NONE then
		
			global.players[global.current_player].left = wrapped_rect.left		
			global.players[global.current_player].top = wrapped_rect.top
			global.players[global.current_player].right = wrapped_rect.right
			global.players[global.current_player].bottom = wrapped_rect.bottom
				
			return COLLISION_WRAPPED
		
		elseif wrap_collision==COLLISION_MOVEABLE_WALL then
			-- Special case to allow Warrior to push moveable wall pieces through
			-- the wrap around.
			if global.current_player==PLAYER_WARRIOR then
				
				handle_moveable_wall(wrap_x, wrap_y, direction)
			end

		end

		return COLLISION_WALL
	end

	return COLLISION_NONE
end

function handle_food(food_type)

	if food_type==FOOD_ALE then
	
		global.players[global.current_player].health = 
			global.players[global.current_player].health + 50

	elseif food_type==FOOD_MEAT then
	
		global.players[global.current_player].health = 
			global.players[global.current_player].health + 100
	end

end

function handle_potion(potion_type)

	local removed = false

	if potion_type==POTION_KILL_ALL_ENEMIES then
	
		for enemy = 1, #global.enemies do
			global.enemies = {}
		end
	
	elseif potion_type==POTION_KILL_DEATH then
	
		for enemy = 1, #global.enemies do
			if global.enemies[enemy].enemy_type==ENEMY_DEATH then
				-- Add temporary marker to signal removal.
				global.enemies[enemy].remove = true
				removed = true
			end
		end
			
	elseif potion_type==POTION_KILL_ENEMIES then
	
		for enemy = 1, #global.enemies do
			if global.enemies[enemy].enemy_type~=ENEMY_DEATH then
				-- Add temporary marker to signal removal.
				global.enemies[enemy].remove = true
				removed = true
			end
		end	
	
	end
	
	if removed then
		remove_dead_enemies()
	end

end

function handle_exit(exit_index, player_index, enemy_index)

	if global.exits[exit_index].state==EXIT_STATE_STABLE then
		-- We're inside a stable / non moving exit.
		local exit_type = global.exits[exit_index]._type
		if exit_type==0 then
			-- Only players can enter a main exit.
			if enemy_index==0 and global.entering_exit==false then
				-- Main exit.
--				dprintf("Entered exit!")

				-- Snap player to the exit location.
				global.entering_exit = true
				global.entering_exit_player = player_index
				global.players[player_index].left = global.exits[exit_index].left
				global.players[player_index].top = global.exits[exit_index].top
				global.players[player_index].right = global.exits[exit_index].left + TILE_WIDTH
				global.players[player_index].bottom = global.exits[exit_index].top + TILE_HEIGHT
				
				play_audio(SOUND_EXIT)
			end
		else
			-- Teleport exit.
			for area = 1, #global.areas do
				if exit_type==global.areas[area].id then
				
					local teleport_ok = can_teleport(area)
					
					if teleport_ok then
						if player_index==0 then
							global.players[enemy_index].left = global.areas[area].left
							global.players[enemy_index].top = global.areas[area].top
							global.players[enemy_index].right = global.areas[area].left + TILE_WIDTH
							global.players[enemy_index].bottom = global.areas[area].top + TILE_HEIGHT												
						else
							global.players[player_index].left = global.areas[area].left
							global.players[player_index].top = global.areas[area].top
							global.players[player_index].right = global.areas[area].left + TILE_WIDTH
							global.players[player_index].bottom = global.areas[area].top + TILE_HEIGHT						
						end
					end
					
					break
				end
			end
		end
	end
end
	
function remove_object(object)

	local map_x = (object.left / TILE_WIDTH) + 1
	local map_y = (object.top / TILE_HEIGHT) + 1
	
	if object.from_trap then

		-- Object was added by a trap.
		-- Go through the trap's tiles and remove the appropriate object.
		for check = 1, #global.traps[object.from_trap].tiles do

			local tx = global.traps[object.from_trap].tiles[check].map_x
			local ty = global.traps[object.from_trap].tiles[check].map_y
			
			if tx==map_x and ty==map_y then

				-- Replace removed object with ground tile.
				global.traps[object.from_trap].tiles[check].tile_category = LEVEL_SPECIAL_RANGE
				global.traps[object.from_trap].tiles[check].tile_index = GRAPHIC_SPECIAL_GROUND
				
				break
			end
		end
	end

	-- Replace object with ground tile.
	global.map[map_y][map_x] = GRAPHIC_SPECIAL_GROUND + LEVEL_SPECIAL_RANGE

end

function remove_door(x, y)

	-- This uses a flood fill style algorithm to remove connected door tiles.
	if y > MAX_LEVEL_HEIGHT or y < 1 or x < 1 or x > MAX_LEVEL_WIDTH then
		-- Out of bounds.
		return
	end
	
	--dprintf("%d %d %d", x, y, global.map[y][x])
	
	if global.map[y][x]>=LEVEL_DOOR_RANGE and global.map[y][x] < LEVEL_EXIT_RANGE then
		
		global.map[y][x] = GRAPHIC_SPECIAL_GROUND + LEVEL_SPECIAL_RANGE

		-- Check the neighbouring tiles for door pieces.
		remove_door(x + 1, y)
		remove_door(x - 1, y)	
		remove_door(x, y - 1)
		remove_door(x, y + 1)
		
	end
end

function update_special_tiles()

	for _exit = 1, #global.exits do
	
		if global.exits[_exit].movement then
			-- Moveable exit.
			-- Is it time to move the exit?			
			if global.exits[_exit].state==EXIT_STATE_STABLE then
				if global.current_time - global.exits[_exit].current_time > global.exits[_exit].timeout then

--					dprintf("Exit timeout EXIT_STATE_STABLE")			
					global.exits[_exit].state = EXIT_STATE_DISAPPEARING
					global.exits[_exit].exit_quad = GRAPHIC_EXIT_APPEAR_4
					global.exits[_exit].frame_time = global.current_time
				end
			
			elseif global.exits[_exit].state==EXIT_STATE_DISAPPEARING then

--				dprintf("Exit timeout EXIT_STATE_DISAPPEARING")
				local index = global.exits[_exit].movement_index
				local map_x = global.exits[_exit].movement[index].map_x
				local map_y = global.exits[_exit].movement[index].map_y

				if global.exits[_exit].exit_quad < GRAPHIC_EXIT_APPEAR_1 then
					-- Replace disappeared exit position with ground tile.
					global.map[map_y][map_x] = GRAPHIC_SPECIAL_GROUND + LEVEL_SPECIAL_RANGE
					global.exits[_exit].state = EXIT_STATE_APPEARING

					local index = global.exits[_exit].movement_index
					-- Work out next exit position.
					index = index + 1
					if index > #global.exits[_exit].movement then
						index = 1
					end
					
					-- Get new position.
					map_x = global.exits[_exit].movement[index].map_x
					map_y = global.exits[_exit].movement[index].map_y
					-- Update exit rectangle.
					global.exits[_exit].left = (map_x - 1) * TILE_WIDTH
					global.exits[_exit].top = (map_y - 1) * TILE_HEIGHT
					global.exits[_exit].right = global.exits[_exit].left + TILE_WIDTH
					global.exits[_exit].bottom = global.exits[_exit].top + TILE_HEIGHT
					
					-- Move on to the next exit position.
					global.exits[_exit].movement_index = index
					global.exits[_exit].exit_quad = GRAPHIC_EXIT_APPEAR_1
					global.exits[_exit].frame_count = 0
	
				else
					if global.current_time - global.exits[_exit].frame_time > EXIT_ANIMATION_TIME then
				
						-- Reset frame timer to current time.
						global.exits[_exit].frame_time = global.current_time

						-- Put disappearing tile at current position.
						global.map[map_y][map_x] = global.exits[_exit].exit_quad + LEVEL_EXIT_RANGE
						-- Move to previous animation step.
						global.exits[_exit].exit_quad = global.exits[_exit].exit_quad - 1							
					end
				end
			
			elseif global.exits[_exit].state==EXIT_STATE_APPEARING then

--				dprintf("Exit timeout EXIT_STATE_APPEARING")
				local index = global.exits[_exit].movement_index
				local map_x = global.exits[_exit].movement[index].map_x
				local map_y = global.exits[_exit].movement[index].map_y	
			
				if global.exits[_exit].exit_quad > GRAPHIC_EXIT_APPEAR_4 then

					-- Put exit tile at new position.
					local tile = global.exits[_exit].movement[index].tile
					global.map[map_y][map_x] = GRAPHIC_EXIT + LEVEL_EXIT_RANGE + tile
					
					global.exits[_exit]._type = tile
					global.exits[_exit].state = EXIT_STATE_STABLE
				
					-- Reset timeout.
					global.exits[_exit].current_time = global.current_time
				else
					if global.current_time - global.exits[_exit].frame_time > EXIT_ANIMATION_TIME then
				
						-- Reset frame timer to current time.
						global.exits[_exit].frame_time = global.current_time

						-- Put disappearing tile at current position.
						global.map[map_y][map_x] = global.exits[_exit].exit_quad + LEVEL_EXIT_RANGE
						-- Move to previous animation step.
						global.exits[_exit].exit_quad = global.exits[_exit].exit_quad + 1
					end
				end				
			end
		end	
	end

end

function get_map_values_from_rect(rect, x_offset, y_offset)

	-- Shrink bounding rectangle by 1 pixel - otherwise we go into the next tile if we're
	-- on a co-ordinate offset +/- 2 pixels away from the next tile.
	local map_rect = {}
	map_rect.left = rect.left
	map_rect.top = rect.top
	map_rect.right = rect.right - 1
	map_rect.bottom = rect.bottom - 1
	
	-- Work out which map values our new rectangle overlaps.
	local m_left = math.floor(map_rect.left / TILE_WIDTH) + 1
	local m_top = math.floor(map_rect.top/ TILE_HEIGHT) + 1
	local m_right = math.floor(map_rect.right / TILE_WIDTH) + 1
	local m_bottom = math.floor(map_rect.bottom/ TILE_HEIGHT) + 1
--	local map_value1 = 0
--	local map_value2 = 0
			
	-- This will be set with the co-ordinates in the map of the area we intersect with.
	local map_coords

	-- Work out from our direction which map indices to get values from.
	if x_offset > 0 then
		if y_offset > 0 then
			-- Down-right.
--			map_value1 = global.map[m_bottom][m_right]
--			map_value2 = global.map[m_bottom][m_left]
			map_coords = {m_bottom, m_right, m_bottom, m_left}
			
		elseif y_offset < 0 then
			-- Up-right.
--			map_value1 = global.map[m_top][m_right]
--			map_value2 = global.map[m_top][m_left]
			map_coords = {m_top, m_right, m_top, m_left}		
		else
			-- Right.
--			map_value1 = global.map[m_top][m_right]
--			map_value2 = global.map[m_bottom][m_right]
			map_coords = {m_top, m_right, m_bottom, m_right}			
		end

	elseif x_offset < 0 then
		if y_offset > 0 then
			-- Down-left.
--			map_value1 = global.map[m_bottom][m_left]
--			map_value2 = global.map[m_bottom][m_right]
			map_coords = {m_bottom, m_left, m_bottom, m_left}
		elseif y_offset < 0 then
			-- Up-left.
--			map_value1 = global.map[m_top][m_left]
--			map_value2 = global.map[m_top][m_right]	
			map_coords = {m_top, m_left, m_top, m_right}
		else
			-- Left.
--			map_value1 = global.map[m_top][m_left]
--			map_value2 = global.map[m_bottom][m_left]
			map_coords = {m_top, m_left, m_bottom, m_left}
		end

	elseif y_offset > 0 then
		-- Down.
--		map_value1 = global.map[m_bottom][m_left]
--		map_value2 = global.map[m_bottom][m_right]
		map_coords = {m_bottom, m_left, m_bottom, m_right}
		
	elseif y_offset < 0 then
		-- Up.
--		map_value1 = global.map[m_top][m_left]
--		map_value2 = global.map[m_top][m_right]
		map_coords = {m_top, m_left, m_top, m_right}
	end

	return map_coords--map_value1, map_value2, map_coords
end

-- Returns COLLISION_NONE if we can move.
function check_movement(x_offset, y_offset, direction)

	local collision = COLLISION_NONE
	
	-- Create new bounding rectangle for the player.
	local player_rect = {}
	player_rect.left = global.players[global.current_player].left + x_offset
	player_rect.top = global.players[global.current_player].top + y_offset
	player_rect.right = player_rect.left + TILE_WIDTH --(TILE_WIDTH - 1)
	player_rect.bottom = player_rect.top + TILE_HEIGHT --(TILE_HEIGHT - 1)
	
	-- See if player has wrapped off screen.
	local wrapped = handle_player_wrapped(player_rect, direction)
	if wrapped==COLLISION_WRAPPED then
		-- They have, so apply new wrapped position.
		return COLLISION_NONE
	elseif wrapped==COLLISION_WALL then
		-- Couldn't wrap player - obstruction? 
		return wrapped
	end
	
	local map_value1 = 0
	local map_value2 = 0
	
	-- This will be set with the co-ordinates in the map of the area we intersect with.
	local map_coords = {}
	
	--map_value1, map_value2, 
	map_coords = get_map_values_from_rect(player_rect,
		x_offset, y_offset)
		
	-- Check static collisions.
	local second_collision = false	
	collision = check_for_tile_collision(map_coords[2], map_coords[1])
	
	local force_second_check = false
	if global.current_player==PLAYER_VALKYRIE and collision==COLLISION_FORCEFIELD then
		-- Ignore this collision.
		force_second_check = true
	end
	
	if collision==COLLISION_NONE or force_second_check then
		collision = check_for_tile_collision(map_coords[4], map_coords[3])
		second_collision = true
	end

	if collision~=COLLISION_NONE then
		-- Handle specific collision logic here.
		if collision==COLLISION_DOOR then

			-- Hit a door.
			if global.players[global.current_player].keys > 0 then
				-- We have a key so remove the door.
				if not second_collision then
					remove_door(map_coords[2], map_coords[1])
				else
					remove_door(map_coords[4], map_coords[3])
				end
				
				global.players[global.current_player].keys = global.players[global.current_player].keys - 1	
				play_audio(SOUND_DOOR_OPEN)
			end
			
		elseif collision==COLLISION_MOVEABLE_WALL then
		
			-- Only the Warrior can move wall pieces.	
			if global.current_player==PLAYER_WARRIOR then
				local moved = false
				if not second_collision then
					moved = handle_moveable_wall(map_coords[2], map_coords[1], direction)
				else					
					moved = handle_moveable_wall(map_coords[4], map_coords[3], direction)
				end
				
				if moved==false then
					collision = COLLISION_WALL
				else
					global.players[PLAYER_WARRIOR].status.custom_action = true
				end			
			end
			
		elseif collision==COLLISION_FORCEFIELD then

			-- Only the Valkyrie can pass through forcefields.
			if global.current_player==PLAYER_VALKYRIE then
				collision = COLLISION_NONE
				global.players[PLAYER_VALKYRIE].status.custom_action = true
			end
			
		end
			
	end
	
	if collision==COLLISION_NONE then
	
		local index
		collision, index = check_for_dynamic_collisions(player_rect, global.current_player, 0)
		-- Handle specific collision logic here.
		if collision==COLLISION_KEY then

			-- Picked up a key.
			global.players[global.current_player].keys = global.players[global.current_player].keys + 1				
			global.level_objects[index].collected = true
			dprintf("Collected key!")
			-- Remove key from level.
			remove_object(global.level_objects[index])
			collision = COLLISION_NONE

			play_audio(SOUND_KEY)

		elseif collision==COLLISION_POTION then

			-- Only the wizard can collect potions.
			if global.current_player==PLAYER_WIZARD then					
				-- Remove potion from level.
				global.level_objects[index].collected = true
				dprintf("Collected potion!")
				handle_potion(global.level_objects[index].potion_type)
				remove_object(global.level_objects[index])
				
				global.players[PLAYER_WIZARD].status.custom_action = true
				
				collision = COLLISION_NONE
			else
				collision = COLLISION_UNCOLLECTABLE_OBJECT
			end
			
		elseif collision==COLLISION_FOOD then

			-- Remove potion from level.
			global.level_objects[index].collected = true
			dprintf("Collected food!")
			handle_food(global.level_objects[index].food_type)
			remove_object(global.level_objects[index])
			collision = COLLISION_NONE
		
		elseif collision==COLLISION_EXIT then
		
			handle_exit(index, global.current_player, 0)
		end
	end
	
	if collision==COLLISION_NONE then
		-- See if we're on a trap tile.
		for trap = 1, #global.traps do

			local in_trap = false
			if global.traps[trap].trap_type~=TRAP_INVALID then
				if rect_intersect_rect(player_rect, global.traps[trap]) then
			
					local left_diff = math.abs(player_rect.left - global.traps[trap].left)
					local top_diff = math.abs(player_rect.top - global.traps[trap].top)
				
					if left_diff <= 4 and top_diff <=4 then
						in_trap = true
					end
				end

	--[[		
				Only handle the trap if the trap doesn't have a player associated with it or the
				trap was activated by this player.
				Otherwise you get the situation where a trap that has been activated by another player, e.g.
				a pressure switch trap gets de-activated when you switch to another player as that player's
				bounding rectangle doesn't intersect.
	--]]		
				if global.traps[trap].player==0 or global.traps[trap].player==global.current_player then
					collision = handle_trap_collisions(trap, in_trap, collision, true)
				end
			end
		end
	end

	return collision
	
end

function check_keys()

	if global.state==STATE_INGAME or global.state==STATE_EDITOR_TEST then
	
		local x_offset = 0
		local y_offset = 0
		local direction = 0
		
		local move = MOVEMENT
		local running = false
		
		if love.keyboard.isDown("lshift") or love.keyboard.isDown("rshift") then
			-- Shift is down so make the player 'run' by doubling the rate of
			-- movement.
 			move = move * 2
			running = true
		else
			global.players[global.current_player].status.running = false
		end

		if love.keyboard.isDown("up") then
			y_offset = -move
			direction = UP
		elseif love.keyboard.isDown("down") then
			y_offset = move
			direction = DOWN
		end
		
		if love.keyboard.isDown("left") then
			x_offset = -move
			if y_offset > 0 then
				direction = DOWN_LEFT
			elseif y_offset < 0 then
				direction = UP_LEFT
			else
				direction = LEFT
			end
			
		elseif love.keyboard.isDown("right") then
			x_offset = move
			if y_offset > 0 then
				direction = DOWN_RIGHT
			elseif y_offset < 0 then
				direction = UP_RIGHT
			else
				direction = RIGHT
			end			
			
		end


		if love.keyboard.isDown("space") then
			
			local fire_direction = direction
			if fire_direction==0 then
				-- Use the last known direction of the player.
				fire_direction = global.players[global.current_player].direction
			end
			
			if handle_weapon(fire_direction) then
				-- Shooting a weapon uses up health, so mark the player as shooting.
				global.players[global.current_player].status.shooting = true
			end
		else
			global.players[global.current_player].status.shooting = false
		end

--		dprintf("%d %d", x_offset, y_offset)

		if direction~=0 then
		
			local collision = check_movement(x_offset, y_offset, direction)
			if collision~=COLLISION_NONE and
				(direction==DOWN_RIGHT or direction==DOWN_LEFT or
					direction==UP_RIGHT or direction==UP_LEFT) then
--[[					
				We hit something whilst we were moving diagonally.
				Try collision test again without upwards / downwards movement.
				Lets us move along walls with diagonal held down instead of
				coming to a dead stop against them.
--]]				
				collision = check_movement(x_offset, 0, direction)
				if collision==COLLISION_NONE then
					y_offset = 0
				else
					collision = check_movement(0, y_offset, direction)
					if collision==COLLISION_NONE then
						x_offset = 0
					end
				end

			end

			if collision~=COLLISION_NONE then
				-- Don't move the player as we collided with something.
				-- We still want to update our direction though.
				x_offset = 0
				y_offset = 0
			end
	
			--dprintf("new_x %d new_y %d", new_x, new_y)
			local new_x = global.players[global.current_player].left + x_offset
			local new_y = global.players[global.current_player].top + y_offset			
			set_player_position(global.current_player, new_x, new_y, direction)
			
			if running then
				-- Running uses up health, so mark the player as running.
				global.players[global.current_player].status.running = true 
			end
		end	
	
	end

end
