
--[[

Puzzlet

editor.lua

(c) bob fossil 2016

--]]

editor_x = 1
editor_y = 1
edit_map = {}
editor_level = 1
editor_category = 1
editor_player_positions = {}
editor_enemy_postions = {}
editor_level_name = ""
editor_exit_data = ""
editor_trap_data = ""
editor_generator_data = ""
editor_health_data = ""


editor_category_lookup = 
	{
	{name = "WALLS", tile = 1},
	{name = "DOORS", tile = 1},
	{name = "EXITS", tile = 1},
	{name = "SPECIALS", tile = 1},
	{name = "PLAYERS", tile = 1},
	{name = "ENEMIES", tile = 1},	
	}

EDITOR_IN_SCREEN = 1
EDITOR_IN_TOOLBOX = 2
	
EDITOR_CATEGORY_WALLS = 1
EDITOR_CATEGORY_DOORS = 2
EDITOR_CATEGORY_EXITS = 3
EDITOR_CATEGORY_SPECIALS = 4
EDITOR_CATEGORY_PLAYERS = 5
EDITOR_CATEGORY_ENEMIES = 6

MAX_CATEGORIES = EDITOR_CATEGORY_ENEMIES

editor_state = EDITOR_IN_SCREEN
editor_test_level = 0
	
tile_description_lookup = 
	{
		-- Walls
		{
		"VERTICAL BOTTOM END",
		"WALL 2",
		"WALL 3",
		"BOTTOM LEFT CORNER",
		"VERTICAL TOP END",
		"VERTICAL MIDDLE",
		"TOP LEFT CORNER",
		"LEFT JUNCTION",
		"WALL 9",
		"BOTTOM RIGHT CORNER",
		"WALL 11",
		"BOTTOM JUNCTION",
		"TOP RIGHT CORNER",
		"RIGHT JUNCTION",
		"TOP JUNCTION",
		"MIDDLE JUNCTION",
		"DESTRUCTABLE 1",
		"DESTRUCTABLE 2",
		"DESTRUCTABLE 3",
		"MOVEABLE",
		"FORCEFIELD",
		},
		-- Doors
		{
		"DOOR 1",
		"DOOR 2", 
		"DOOR 3",
		"DOOR 4",
		"DOOR 5",
		"DOOR 6",
		"DOOR 7",
		"DOOR 8",
		"DOOR 9",
		"DOOR 10",
		"DOOR 11",
		"DOOR 12",
		"DOOR 13",
		"DOOR 14",
                "DOOR 15",
		},
		-- Exits
		{
		"EXIT",
		"EXIT TO 1",
		"EXIT TO 2",
		"EXIT TO 3",
		"EXIT TO 4",
		"EXIT TO 5",
		"EXIT TO 6",
		"EXIT TO 7",
		"EXIT TO 8",
		},
		-- Specials,
		{
		"GROUND",
		"AREA 1",
		"AREA 2",
		"AREA 3",
		"AREA 4",
		"AREA 5",
		"AREA 6",
		"AREA 7",
		"AREA 8",
		"KEY",
		"POTION",
		"BONES",
		"GENERATOR DAMAGED",
		"GENERATOR OFF",
		"GENERATOR ON",
		"FORCEFIELD",
		"TRAP",
		"ALE",
		"MEAT",
		},
		-- Players
		{
		"WARRIOR",
		"VALKYRIE",
		"WIZARD",
		"ELF",
		},
		-- Enemies
		{
		"DEATH",
		"GHOST",
		"DEMON",
		},
	}

function init_editor()

	global.state = STATE_EDITOR
	
	editor_category_lookup[EDITOR_CATEGORY_WALLS].gfx = GRAPHIC_WALLS
	editor_category_lookup[EDITOR_CATEGORY_WALLS].range = LEVEL_WALL_RANGE
	editor_category_lookup[EDITOR_CATEGORY_WALLS].max_tile = MAX_GRAPHIC_WALL
		
	editor_category_lookup[EDITOR_CATEGORY_DOORS].gfx = GRAPHIC_DOORS
	editor_category_lookup[EDITOR_CATEGORY_DOORS].range = LEVEL_DOOR_RANGE
	editor_category_lookup[EDITOR_CATEGORY_DOORS].max_tile = MAX_GRAPHIC_DOOR

	editor_category_lookup[EDITOR_CATEGORY_EXITS].gfx = GRAPHIC_EXITS
	editor_category_lookup[EDITOR_CATEGORY_EXITS].range = LEVEL_EXIT_RANGE
	editor_category_lookup[EDITOR_CATEGORY_EXITS].max_tile = GRAPHIC_EXIT_8

	editor_category_lookup[EDITOR_CATEGORY_SPECIALS].gfx = GRAPHIC_SPECIALS
	editor_category_lookup[EDITOR_CATEGORY_SPECIALS].range = LEVEL_SPECIAL_RANGE
	editor_category_lookup[EDITOR_CATEGORY_SPECIALS].max_tile = MAX_GRAPHIC_SPECIAL

	editor_category_lookup[EDITOR_CATEGORY_PLAYERS].gfx = GRAPHIC_PLAYERS
	editor_category_lookup[EDITOR_CATEGORY_PLAYERS].range = LEVEL_PLAYER_RANGE
	editor_category_lookup[EDITOR_CATEGORY_PLAYERS].custom_tiles = {GRAPHIC_WARRIOR_DOWN_1, GRAPHIC_VALKYRIE_DOWN_1,
		GRAPHIC_WIZARD_DOWN_1, GRAPHIC_ELF_DOWN_1}
	editor_category_lookup[EDITOR_CATEGORY_PLAYERS].max_tile = MAX_PLAYERS

	editor_category_lookup[EDITOR_CATEGORY_ENEMIES].gfx = GRAPHIC_ENEMIES
	editor_category_lookup[EDITOR_CATEGORY_ENEMIES].range = LEVEL_ENEMY_RANGE
	editor_category_lookup[EDITOR_CATEGORY_ENEMIES].custom_tiles = {GRAPHIC_DEATH_DOWN_1, GRAPHIC_GHOST_DOWN_1, GRAPHIC_DEMON_DOWN_1}
	editor_category_lookup[EDITOR_CATEGORY_ENEMIES].max_tile = MAX_ENEMIES
	
	editor_test_level = level_count + 1

	get_edit_map()
end

function get_edit_map()

	reset_players()
	
	edit_map = {}
	edit_map = create_map_from_level_data(editor_level, true)
	
	editor_level_name = global.level_data[editor_level].name
	editor_exit_data = global.level_data[editor_level].exits
	editor_trap_data = global.level_data[editor_level].traps
	editor_generator_data = global.level_data[editor_level].generators
	editor_health_data = global.level_data[editor_level].health_data

	editor_player_positions = {}
	local editor_player_positions_size = 0
			
	for add = 1, MAX_PLAYERS do
	
		if global.players[add].active then
		
			-- Add player to the edit map.
			local map_x = (global.players[add].left / TILE_WIDTH) + 1
			local map_y = (global.players[add].top / TILE_HEIGHT) + 1
			edit_map[map_y][map_x] = LEVEL_PLAYER_RANGE + add
			
			-- Make a note of the player so we don't add them again.
			editor_player_positions_size = editor_player_positions_size + 1
			editor_player_positions[editor_player_positions_size] =
				{x = map_x, y = map_y, player = add + LEVEL_PLAYER_RANGE}
		end
	end

end

function clear_edit_map()

	for y = 1, MAX_LEVEL_HEIGHT do
		edit_map[y]={}
		for x= 1, MAX_LEVEL_WIDTH do
			edit_map[y][x] = LEVEL_SPECIAL_RANGE + GRAPHIC_SPECIAL_GROUND			
		end
	end	

end

function save_edit_map(map_string)

	local dir = love.filesystem.getSaveDirectory( )
	dprintf("Saving level to '%s'", dir)
	
	love.filesystem.write("editor.dat", map_string, #map_string)
end

function place_tile(override)

	local value
	if override then
		value = override
	else
		value = editor_category_lookup[editor_category].range + 
			editor_category_lookup[editor_category].tile
	end

	if editor_category==EDITOR_CATEGORY_PLAYERS then

		-- Add a player.
		local already_exists = false
		local editor_player_positions_size = #editor_player_positions
		
		if editor_player_positions_size > 0 then
			-- Don't add the player if they've already been added.

			for check = 1, editor_player_positions_size do

				if editor_player_positions[check].player==value then

					already_exists = true
					break
				end
			end
		end
		
		if not already_exists then
			-- Just add the player.
			editor_player_positions[editor_player_positions_size + 1] =
				{x = editor_x, y = editor_y, player = value}
				
			edit_map[editor_y][editor_x] = value
		end

	else
		edit_map[editor_y][editor_x] = value
		
		local editor_player_positions_size = #editor_player_positions
		if editor_player_positions_size > 0 then
			-- Remove player if they've been overwritten by a block.
			local remove_index = 0
			for check = 1, editor_player_positions_size do
			
				if editor_player_positions[check].x==editor_x and
					editor_player_positions[check].y==editor_y then
					-- Remove this player.
					remove_index = check
					break
				end
			end
			
			if remove_index~=0 then
				table.remove(editor_player_positions, remove_index)
			end
		end
	end


end

function handle_editor_keypress(key)

	local control_down = false
	if love.keyboard.isDown("lctrl") or love.keyboard.isDown("rctrl") then
	
		control_down = true
	end

	if key=="left" then
		
		if editor_state == EDITOR_IN_SCREEN then
	
			if editor_x > 1 then
				editor_x = editor_x - 1
			end
		elseif editor_state == EDITOR_IN_TOOLBOX then
		
			if editor_category > 1 then
				editor_category = editor_category - 1
			end
		end
	
	elseif key=="right" then

		if editor_state == EDITOR_IN_SCREEN then

			if editor_x < MAX_LEVEL_WIDTH then
				editor_x = editor_x + 1
			end
		elseif editor_state == EDITOR_IN_TOOLBOX then
		
			if editor_category < MAX_CATEGORIES then
				editor_category = editor_category + 1
			end
		end

	elseif key=="up" then

		if editor_state == EDITOR_IN_SCREEN then
			if editor_y > 1 then
				editor_y = editor_y - 1
			end
		elseif editor_state == EDITOR_IN_TOOLBOX then
		
			if editor_category_lookup[editor_category].tile > 1 then
				editor_category_lookup[editor_category].tile =
					editor_category_lookup[editor_category].tile - 1
			end
		end
	
	elseif key=="down" then

		if editor_state == EDITOR_IN_SCREEN then
			if editor_y < MAX_LEVEL_HEIGHT then
				editor_y = editor_y + 1
			end
		elseif editor_state == EDITOR_IN_TOOLBOX then

			if editor_category_lookup[editor_category].tile < 
				editor_category_lookup[editor_category].max_tile then
				editor_category_lookup[editor_category].tile =
					editor_category_lookup[editor_category].tile + 1
			end
				
		end
		
	elseif key=="c" then

		if editor_state == EDITOR_IN_SCREEN then

			-- Clone the current tile.
			local tile = edit_map[editor_y][editor_x]
			local start = 0
			-- Find the category the screen tile belongs to.
			for category = EDITOR_CATEGORY_WALLS, EDITOR_CATEGORY_ENEMIES do
				
				if tile > start and tile < start + LEVEL_RANGE_SIZE then
							
					-- Set the editor category and tile.
					editor_category = category				
					editor_category_lookup[editor_category].tile = tile - start
					break
				end
				
				start = start + LEVEL_RANGE_SIZE
			end
		end
		
	elseif key=="d" then

		local map_string = map_to_string(edit_map, true)
		dprintf(map_string)

	elseif key=="l" then
	
		editor_level = editor_level + 1
		if editor_level > global.levels then
			editor_level = 1
		end
		
		get_edit_map()		
		
	elseif key=="r" then

		-- Reload levels.dat.
		load_levels()
		
		dprintf("Loaded %d levels.", #global.level_data)
		
		get_edit_map()
		
	elseif key=="s" then

		local map_string = map_to_string(edit_map, true)
		save_edit_map(map_string)
		
	elseif key=="t" then
		-- Run the current level from the Editor.
		local level_data = map_to_string(edit_map, false)
		local health_data = editor_health_data
		if control_down then
			-- Test with larger health values.
			health_data = "999,999,999,999"
		end
		
		if global.level_data[editor_test_level] then
			table.remove(global.level_data, editor_test_level)
		end
		
		add_level(level_data, editor_level_name, editor_exit_data, editor_trap_data,
			editor_generator_data, health_data, true)
		init_level(editor_test_level)
		
		-- Test the level
		global.level = editor_test_level
		global.state = STATE_EDITOR_TEST
		
	elseif key=="tab" then
		
		if editor_state == EDITOR_IN_SCREEN then
			editor_state = EDITOR_IN_TOOLBOX
		elseif editor_state == EDITOR_IN_TOOLBOX then
			editor_state = EDITOR_IN_SCREEN
		end
		
	elseif key=="space" then
	
		if editor_state == EDITOR_IN_SCREEN then
		
			place_tile()
		else
		end
		
	elseif key=="escape" then
	
		if global.level_data[editor_test_level] then
			table.remove(global.level_data, editor_test_level)
		end
	
		-- Switch back to the intro screen.
		init_titlescreen()

	elseif key=="delete" then
		-- Erase the current tile.
		place_tile(LEVEL_SPECIAL_RANGE + GRAPHIC_SPECIAL_GROUND)		
		
	elseif key=="home" then
		
		if editor_state == EDITOR_IN_SCREEN then

			if control_down then
				-- Go to the first square.
				editor_x = 1
				editor_y = 1
			else
				-- Go to the first column in this row.
				editor_x = 1
			end
		
		elseif editor_state == EDITOR_IN_TOOLBOX then

			-- Go to first item in category.
			editor_category_lookup[editor_category].tile = 1

		end
	
	elseif key=="end" then

		if editor_state == EDITOR_IN_SCREEN then

			if control_down then
				-- Go to the last square.
				editor_x = MAX_LEVEL_WIDTH
				editor_y = MAX_LEVEL_HEIGHT
			else
				-- Go to the last column in this row.
				editor_x = MAX_LEVEL_WIDTH
			end

		elseif editor_state == EDITOR_IN_TOOLBOX then

			-- Go to last item in category.
			editor_category_lookup[editor_category].tile =
				editor_category_lookup[editor_category].max_tile

		end

	end
	
end

function draw_editor_level()

	local map_height = #edit_map
	
	local map_width
	local map_value
	local x = 0
	local y = 0
	
	-- Values in the map are range bound.
	for index = 1, map_height do
		map_width = #edit_map[index]
		for width = 1, map_width do

			map_value = edit_map[index][width]
			
			if map_value < LEVEL_DOOR_RANGE then
				love.graphics.draw(tile_graphic, wall_graphic_quads[map_value], x, y)
			elseif map_value < LEVEL_EXIT_RANGE then
				love.graphics.draw(tile_graphic, door_graphic_quads[map_value - LEVEL_DOOR_RANGE], x, y)				
			elseif map_value < LEVEL_SPECIAL_RANGE then
				love.graphics.draw(tile_graphic, exit_graphic_quads[map_value - LEVEL_EXIT_RANGE], x, y)
			elseif map_value < LEVEL_PLAYER_RANGE then
				love.graphics.draw(tile_graphic, special_graphic_quads[map_value - LEVEL_SPECIAL_RANGE], x, y)
			elseif map_value < LEVEL_ENEMY_RANGE then
				local quad = editor_category_lookup[EDITOR_CATEGORY_PLAYERS].custom_tiles[map_value - LEVEL_PLAYER_RANGE]						
				love.graphics.draw(player_graphic, player_graphic_quads[quad], x, y)
			else
				local quad = editor_category_lookup[EDITOR_CATEGORY_ENEMIES].custom_tiles[map_value - LEVEL_ENEMY_RANGE]			
				love.graphics.draw(enemy_graphic, enemy_graphic_quads[quad], x, y)
			end
			
			x = x + TILE_WIDTH
		end
		
		x = 0
		y = y + TILE_HEIGHT
	end

end

function draw_editor()

	love.graphics.setColor(255, 255, 255, 255)

	-- Print level name.
	draw_text(-1, SCREEN_HEIGHT - (FONT_CHARACTER_HEIGHT * 4), editor_level_name, 255, 255, 255)
	
	local status_string1 = string.format("%s %s", editor_category_lookup[editor_category].name,
		tile_description_lookup[editor_category][editor_category_lookup[editor_category].tile])
	draw_text(0, SCREEN_HEIGHT - (FONT_CHARACTER_HEIGHT * 3), status_string1, 255, 255, 255)
	
	local status_string2 = string.format("X:%02d Y:%02d L:%02d", editor_x, editor_y, editor_level)
	draw_text(SCREEN_WIDTH / 2, SCREEN_HEIGHT - 16, status_string2, 255, 255, 255)
	
	draw_editor_level()
	
	local draw_x = 0
	local draw_y = 0
	local ex = (editor_x - 1) * TILE_WIDTH
	local ey = (editor_y - 1) * TILE_HEIGHT
	
	for y = 1, MAX_LEVEL_HEIGHT do

		for x = 1, MAX_LEVEL_WIDTH do

			if editor_state==EDITOR_IN_SCREEN then
				if draw_x==ex and draw_y==ey then
					love.graphics.setColor(255, 255, 0, 255)
					love.graphics.rectangle("line", draw_x + 1, draw_y + 1, TILE_WIDTH - 2, TILE_HEIGHT -2)
					
				else
					love.graphics.setColor(128, 128, 128, 255)
				end
			else
				love.graphics.setColor(128, 128, 128, 255)	
			end

			love.graphics.rectangle("line", draw_x, draw_y, TILE_WIDTH, TILE_HEIGHT)
			
			draw_x = draw_x + TILE_WIDTH
		end
		draw_x = 0
		draw_y = draw_y + TILE_HEIGHT
		
	end

	love.graphics.setColor(255, 255, 255, 255)
		
	local category_x = 0
	for category = 1, MAX_CATEGORIES do

		if editor_category_lookup[category].custom_tiles then
		
			local tile = editor_category_lookup[category].tile		
			draw_graphic(editor_category_lookup[category].custom_tiles[tile], category_x,
				SCREEN_HEIGHT - TILE_WIDTH, editor_category_lookup[category].gfx)		
		else
			draw_graphic(editor_category_lookup[category].tile, category_x,
				SCREEN_HEIGHT - TILE_WIDTH, editor_category_lookup[category].gfx)
		end
		category_x = category_x + TILE_WIDTH
	end

	-- Draw category highlight depending on the state we're in.
	category_x = (editor_category - 1) * TILE_WIDTH	
	if editor_state==EDITOR_IN_SCREEN then
		love.graphics.setColor(255, 255, 255, 255)
	else
		love.graphics.setColor(255, 255, 0, 255)
	end
	love.graphics.rectangle("line", category_x, SCREEN_HEIGHT - TILE_WIDTH, TILE_WIDTH, TILE_HEIGHT)
end