
--[[

Puzzlet

enemies.lua

(c) bob fossil 2016

--]]

ENEMY_DEATH = 1
ENEMY_GHOST = 2
ENEMY_DEMON = 3
ENEMY_FIREBALL = 4
MAX_ENEMIES = 3

--[[
	Generators 'G' potentially have 8 slots around them which enemies can be spawned from.

	1	2	3
	
	8	G	4
	
	7	6	5

--]]

spawn_slots = 
	{
	{x = -1, y = -1, direction = UP_LEFT},	-- Slot 1
	{x = 0, y = -1, direction = UP},	-- Slot 2
	{x = 1, y = -1, direction = UP_RIGHT},	-- Slot 3
	{x = 1, y = 0, direction = RIGHT},	-- Slot 4
	{x = 1, y = 1, direction = DOWN_RIGHT},	-- Slot 5
	{x = 0, y = 1, direction = DOWN},	-- Slot 6
	{x = -1, y = 1, direction = DOWN_LEFT}, -- Slot 7
	{x = -1, y =0, direction = LEFT}, 	-- Slot 8
	}
	
MAX_SLOTS = 8

require "graphics_constants"

enemy_defaults=
	{
	-- Death
	{first_quad = GRAPHIC_DEATH_UP_1 - 1, direction = DOWN},
	-- Ghost
	{first_quad = GRAPHIC_GHOST_UP_1 - 1, direction = DOWN},
	-- Demon
	{first_quad = GRAPHIC_DEMON_UP_1 - 1, direction = DOWN},
	-- Fireball
	{first_quad = GRAPHIC_FIREBALL_UP_1 - 1, direction = DOWN},	
	}

function add_enemy(index, x, y, direction)

	if global.debug_mode then
		dprintf("Added enemy type %d at %d %d", index, x, y)
	end

	local enemies = #global.enemies
	local enemy_table = {left = x, top = y, right = x + TILE_WIDTH, bottom = y + TILE_HEIGHT, enemy_type = index,
		first_quad = enemy_defaults[index].first_quad, direction = enemy_defaults[index].direction, frame = 0}
	
	if direction~=0 then
		enemy_table.direction = direction
	end
	
	global.enemies[enemies + 1] = enemy_table

end

function set_enemy_position(enemy_index, new_x, new_y, new_direction)

	--dprintf("nd %d d %d", new_direction, global.players[player_index].direction)
	local moved = false
	if new_x~=global.enemies[enemy_index].left or new_y~=global.enemies[enemy_index].top then
		moved = true
	end

	if new_direction~=global.enemies[enemy_index].direction then
				
		global.enemies[enemy_index].frame = 0
		global.enemies[enemy_index].direction = new_direction
		
	else
		-- Only animate if we moved.
		if global.update_frame then --moved
			global.enemies[enemy_index].frame = global.enemies[enemy_index].frame + 1
			if global.enemies[enemy_index].frame > MAX_FRAMES then
				global.enemies[enemy_index].frame = 0
			end			
		end
		
	end
	
	--dprintf("f %d - %d", global.players[player_index].frame, global.players[player_index].first_quad)
	
	if moved then
		global.enemies[enemy_index].left = new_x
		global.enemies[enemy_index].top = new_y
		global.enemies[enemy_index].right = new_x + TILE_WIDTH
		global.enemies[enemy_index].bottom = new_y + TILE_HEIGHT
	end
	
end

function remove_dead_enemies()

	local enemies = #global.enemies
	for index = enemies, 1, -1 do	
		if global.enemies[index].remove then
			table.remove(global.enemies, index)
		end
	end

end

function check_enemy_collision(enemy, x_offset, y_offset)

	local map_value1 = 0
	local map_value2 = 0
		
	local collision = COLLISION_NONE
	
	local enemy_rect = {}
	enemy_rect.left = global.enemies[enemy].left + x_offset
	enemy_rect.top = global.enemies[enemy].top + y_offset
	enemy_rect.right = enemy_rect.left + TILE_WIDTH --(TILE_WIDTH - 1)
	enemy_rect.bottom = enemy_rect.top + TILE_HEIGHT --(TILE_HEIGHT - 1)	
	
	-- This will be set with the co-ordinates in the map of the area we intersect with.
	local map_coords = {}
	
	--map_value1, map_value2,
	map_coords = get_map_values_from_rect(enemy_rect,
		x_offset, y_offset)
	
	-- Check static collisions.	
	collision = check_for_tile_collision(map_coords[2], map_coords[1])
	if collision == COLLISION_NONE then
		collision = check_for_tile_collision(map_coords[4], map_coords[3])
	end

	-- Check other enemies.
	if collision == COLLISION_NONE then
	
		local index
		collision, index = check_for_dynamic_collisions(enemy_rect, 0, enemy)
		-- Handle specific collision logic here.
		if collision==COLLISION_PLAYER then

			if global.enemies[enemy].enemy_type==ENEMY_DEATH and
				global.players[index].alive then
				-- Contact with Death is fatal.
				global.players[index].alive = false
				dprintf("Player %d is dead!", index)
				-- Switch to the next player.
				if index==global.current_player then
					global.current_player = get_next_player(global.current_player)
				end
			end

		elseif collision==COLLISION_EXIT then
		
			handle_exit(index, 0, enemy)

		end
	end
	
	--dprintf("cec %d %d x %d y %d col %d mt %d m1 %d m2 %d", x_offset, y_offset, enemy_rect.left, enemy_rect.top, collision, map_total, map_value1, map_value2)
		
	return collision	
end

function find_nearest_player(enemy)

	local distance = 65536
	local nearest = 0
	
	local enemy_direction = 0	
	local x_offset = 0
	local y_offset = 0	

	-- Find the nearest player.
	for player = 1, #global.players do
		if global.players[player].active==true and
			global.players[player].alive==true and
			global.players[player].exited==false then
			
			local dx = math.abs(global.players[player].left -
				global.enemies[enemy].left)
			local dy = math.abs(global.players[player].top -
				global.enemies[enemy].top)

			local new_distance = math.sqrt((dx ^ 2) + (dy ^ 2))

			--dprintf("player %d %d %d (%d)", player, dx, dy, new_distance)
					
			if new_distance < distance then
				distance = new_distance
				nearest = player
			end
		end
	end
	
	
	if nearest~=0 then

		--dprintf("Nearest player %d", nearest)
	
		-- Work out how best to move towards them.
		local enemy_x = global.enemies[enemy].left 
		local enemy_y = global.enemies[enemy].top 

		if global.players[nearest].left < enemy_x then
			x_offset = -MOVEMENT
		elseif global.players[nearest].left > enemy_x then
			x_offset = MOVEMENT
		end
		
		if global.players[nearest].top < enemy_y then
			y_offset = -MOVEMENT
		elseif global.players[nearest].top > enemy_y then
			y_offset = MOVEMENT
		end				
		
		if x_offset < 0 then
		
			if y_offset < 0 then
				enemy_direction = UP_LEFT
			elseif y_offset > 0 then
				enemy_direction = DOWN_LEFT
			else 
				enemy_direction = LEFT
			end
		elseif x_offset > 0 then
		
			if y_offset < 0 then
				enemy_direction = UP_RIGHT
			elseif y_offset > 0 then
				enemy_direction = DOWN_RIGHT
			else
				enemy_direction = RIGHT
			end
		else
			if y_offset < 0 then
				enemy_direction = UP
			elseif y_offset > 0 then
				enemy_direction = DOWN
			end
		end
	end
	
	return enemy_direction, x_offset, y_offset

end

function update_generators()

	for index = 1, #global.generators do
	
		local generator = global.generators[index]
		if generator.state==GENERATOR_ON and
			global.current_time - generator.current_time > generator.spawn_time then
		
			--dprintf("Attempting to spawn from generator %d", index)
			
			if generator.spawned <= generator.spawn_count then

				local map_x = (generator.left / TILE_WIDTH) + 1
				local map_y = (generator.top / TILE_HEIGHT) + 1
				
				local spawn_x
				local spawn_y
				local spawn_rect={}

				-- Pass 1. Work out which slots are available.
				for check = 1, MAX_SLOTS do

					generator.slot_table[check] = false

					-- Find a free map tile to spawn to.
					spawn_x = map_x + spawn_slots[check].x
					spawn_y = map_y + spawn_slots[check].y
					
					if check_for_tile_collision(spawn_x, spawn_y)==COLLISION_NONE then
						
						-- Map is clear. Check for players / enemies.
						spawn_rect.left = (spawn_x - 1) * TILE_WIDTH
						spawn_rect.top = (spawn_y - 1) * TILE_HEIGHT
						spawn_rect.right = spawn_rect.left + (TILE_WIDTH - 1)
						spawn_rect.bottom = spawn_rect.top + (TILE_HEIGHT - 1)
						
						if check_for_dynamic_collisions(spawn_rect, 0, 0)==COLLISION_NONE then
							-- Found a free slot to generate from.
							generator.slot_table[check] = true
--						else
--							dprintf("slot %d dynamic collision failed", check)							
						end
--					else
--					 	dprintf("slot %d tile collision failed", check)
					end
				
				end
				
				-- We try and generate from the next slot after the last one we generated from.
				local current_slot = 1
				local count = 1
				if generator.next_slot~=0 then
					current_slot = generator.next_slot
				end
				
				-- Pass 2. Go through all the slots and find the next free one.
				while count <= MAX_SLOTS do
				
					local next_slot = current_slot + 1
					if next_slot > MAX_SLOTS then
						-- Go back to the first slot.
						next_slot = 1
					end
					
					if generator.slot_table[current_slot] == true then

						-- Work out where to put our new enemy.
						spawn_x = ((map_x + spawn_slots[current_slot].x) - 1) * TILE_WIDTH
						spawn_y = ((map_y + spawn_slots[current_slot].y) - 1) * TILE_HEIGHT
						
						add_enemy(generator.enemy_type, spawn_x, spawn_y, spawn_slots[current_slot].direction)

						if generator.spawn_count > 0 then
							generator.spawned = generator.spawned + 1
						end
						
						generator.next_slot = next_slot
						
						break
					end
				
					count = count + 1
					current_slot = next_slot
				end
			
			end
			generator.current_time = global.current_time
		end
	
	end

end

function update_enemies()

	update_generators()

	for enemy = 1, #global.enemies do

		local enemy_direction
		local x_offset
		local y_offset
		local enemy_direction, x_offset, y_offset = find_nearest_player(enemy)

		if enemy_direction~=0 then
			local collision = check_enemy_collision(enemy, x_offset, y_offset)
			if collision~=COLLISION_NONE and
				(enemy_direction==UP_LEFT or enemy_direction==DOWN_LEFT or 
				enemy_direction==UP_RIGHT or enemy_direction==DOWN_RIGHT) then

--[[					
				We hit something whilst we were moving diagonally.
				Try collision test again without upwards / downwards movement.
--]]						
		
				collision = check_enemy_collision(enemy, x_offset, 0)
				--dprintf("col1 %d", collision)
				
				if collision==COLLISION_NONE then
					y_offset = 0
					if enemy_direction==UP_LEFT then
						enemy_direction = LEFT
					elseif enemy_direction==UP_RIGHT then
						enemy_direction = RIGHT
					elseif enemy_direction==DOWN_LEFT then
						enemy_direction = LEFT
					elseif enemy_direction==DOWN_RIGHT then
						enemy_direction = RIGHT
					end
					
				else
					collision = check_enemy_collision(enemy, 0, y_offset)
					--dprintf("col2 %d", collision)
					
					if collision==COLLISION_NONE then
						x_offset = 0
						if enemy_direction==UP_LEFT then
							enemy_direction = UP
						elseif enemy_direction==UP_RIGHT then
							enemy_direction = UP
						elseif enemy_direction==DOWN_LEFT then
							enemy_direction = DOWN
						elseif enemy_direction==DOWN_RIGHT then
							enemy_direction = DOWN
						end								
					end
				end
			end
								
			if collision==COLLISION_NONE then
				set_enemy_position(enemy, global.enemies[enemy].left + x_offset,
					global.enemies[enemy].top + y_offset, enemy_direction)
			else
				-- Don't move, but update direction.
				set_enemy_position(enemy, global.enemies[enemy].left,
					global.enemies[enemy].top, enemy_direction)
			end
				
		end			
	end

end

