
function handle_game_keypress(key)

	if key=="s" then
		-- Swap to the next active player.
		if global.entering_exit==false then
			global.current_player = get_next_player(global.current_player)
		end
	-- Player 'speed dial' selection keys.
	elseif key=="1" then
		if can_switch_to_player(PLAYER_WARRIOR) then
			global.current_player = PLAYER_WARRIOR
		end
	elseif key=="2" then
		if can_switch_to_player(PLAYER_VALKYRIE) then
			global.current_player = PLAYER_VALKYRIE
		end
	elseif key=="3" then
		if can_switch_to_player(PLAYER_WIZARD) then
			global.current_player = PLAYER_WIZARD
		end
	elseif key=="4" then
		if can_switch_to_player(PLAYER_ELF) then
			global.current_player = PLAYER_ELF
		end			
	elseif key=="escape" then
		if global.state==STATE_EDITOR_TEST then
			global.state = STATE_EDITOR
		else
			-- Display 'Retry / Exit message.
			set_message(MESSAGE_RESTART_QUIT_LEVEL, {restart_level, quit_level})
		end
--	elseif key=="d" then
--		debug.debug()
	end

end

-- Handle logic for the game state.
function update_game()

	if global.current_time - global.last_health_time > PLAYER_HEALTH_TIME_PERIOD then
		-- Decrease players health every second they're active.
		if not global.entering_exit then

			global.players[global.current_player].health =
				global.players[global.current_player].health - 1

			if global.players[global.current_player].status.running then
				-- Running uses up health.
				global.players[global.current_player].health =
					global.players[global.current_player].health - 1
			end
			
			if global.players[global.current_player].status.shooting then
				-- Shooting weapons uses up health.
				global.players[global.current_player].health =
					global.players[global.current_player].health - 1
			end

			if global.players[global.current_player].status.custom_action then
				-- Doing a player's custom action, e.g. pushing a block, magic
				-- uses up some health.
				global.players[global.current_player].health =
					global.players[global.current_player].health - 1
				
				global.players[global.current_player].status.custom_action = false
			end
			
			if global.players[global.current_player].health < 0 then
			
				-- Player has run out of health, so kill them off.
				global.players[global.current_player].health = 0
				global.players[global.current_player].alive = false
				dprintf("Player %d ran out of health!", global.current_player)
				global.current_player = get_next_player(global.current_player)
			
			elseif global.players[global.current_player].health <= PLAYER_LOW_HEALTH and
				(global.players[global.current_player].health % 2 == 0) then
				-- Play low health sound every other decrement.
				play_audio(SOUND_LOW_HEALTH)
			end
		end
			
		global.last_health_time = global.current_time
	end

	local result = check_players()
	
	if result==PLAYERS_ALL_EXITED then
		-- Tell the player they've finished the level.
		set_message(MESSAGE_NEXT_LEVEL, {next_level})
		
	elseif result==PLAYERS_SOME_EXITED then
		-- Tell the player they need to restart.
		set_message(MESSAGE_RESTART_SOME_DEAD, {restart_level})

	elseif result==PLAYERS_ALL_DEAD then
		-- Tell the player they need to restart.
		set_message(MESSAGE_RESTART_ALL_DEAD, {restart_level})
	else
		-- Carry on the game.
		update_special_tiles()
		update_weapons()
		check_keys()
		update_enemies()
	end
end

-- Draws the main game screen.
function draw_game()

	love.graphics.setColor(255, 255, 255, 255)

	draw_level()
	draw_players()
	draw_weapons()
	draw_enemies()

	-- Print level name.
	draw_text(-1, SCREEN_HEIGHT - FONT_CHARACTER_HEIGHT, get_level_name(), 255, 255, 255)
	if global.current_player~=0 then
	
		local player = global.players[global.current_player]
	
		-- Displayer active player and their stats.
		draw_text(FONT_CHARACTER_WIDTH, SCREEN_HEIGHT - (FONT_CHARACTER_HEIGHT * 3), player.name,
			player.r, player.g, player.b)

--		local stats = string.format("KEYS: %02d HEALTH: %03d", global.players[global.current_player].keys, global.players[global.current_player].health)
			
		local key_stats = string.format("KEYS: %02d", player.keys)
		draw_text(SCREEN_WIDTH / 2, SCREEN_HEIGHT - (FONT_CHARACTER_HEIGHT * 3), key_stats, 255, 255, 255)
		local health_stats = string.format("HEALTH: %03d", player.health)
		if player.health > PLAYER_LOW_HEALTH then
			draw_text((SCREEN_WIDTH / 2) + FONT_CHARACTER_WIDTH * 9, SCREEN_HEIGHT - (FONT_CHARACTER_HEIGHT * 3), health_stats, 255, 255, 255)
		else
			-- Draw low health in red.
			draw_text((SCREEN_WIDTH / 2) + FONT_CHARACTER_WIDTH * 9, SCREEN_HEIGHT - (FONT_CHARACTER_HEIGHT * 3), health_stats, 255, 0, 0)
		end
	end
	
	-- Do any debug drawing stuff here.
	if global.debug_mode then
--[[	
		for key = 1, #global.keys do	
			love.graphics.rectangle("line", global.keys[key].l, global.keys[key].t, global.keys[key].r - global.keys[key].l, global.keys[key].b - global.keys[key].t)
		end	
		for _exit = 1, #global.exits do	
			love.graphics.rectangle("line", global.exits[_exit].l, global.exits[_exit].t, global.exits[_exit].r - global.exits[_exit].l, global.exits[_exit].b - global.exits[_exit].t)
		end	

		if global.current_player~=0 then
			love.graphics.setColor(global.players[global.current_player].r, global.players[global.current_player].g, global.players[global.current_player].b, 255)	
			love.graphics.rectangle("line", global.players[global.current_player].x, global.players[global.current_player].y, TILE_WIDTH, TILE_HEIGHT)
		end
--]]		
	end
	
end
