
--[[

graphics.lua

(c) bob fossil 2015-2016

--]]

require "constants"
require "graphics_constants"

--
tile_graphic = nil
wall_graphic_quads = {}
door_graphic_quads={}
exit_graphic_quads={}
special_graphic_quads = {}

enemy_graphic = nil
enemy_graphic_quads={}

player_graphic = nil
player_graphic_quads = {}


GRAPHIC_WALLS = 1
GRAPHIC_DOORS = 2
GRAPHIC_EXITS = 3
GRAPHIC_SPECIALS = 4
GRAPHIC_ENEMIES = 5
GRAPHIC_PLAYERS = 6

graphics_table = {}

function init_graphics()

	tile_graphic = love.graphics.newImage("tiles.png")
	local width = tile_graphic:getWidth()
	local height = tile_graphic:getHeight()
	local y = 0
	local x = 0

	-- Setup wall tiles.
	for sprite = 1, MAX_GRAPHIC_WALL do
		wall_graphic_quads[sprite] = love.graphics.newQuad(x, GRAPHIC_WALL_OFFSET, TILE_WIDTH, TILE_HEIGHT, width, height)
			
			--dprintf("player (%d) %d %d", sprite, x, y)
			
			x = x + TILE_WIDTH
			
		end

	-- Setup door tiles.
	x = 0
	for sprite = 1, MAX_GRAPHIC_DOOR do
		door_graphic_quads[sprite] = love.graphics.newQuad(x, GRAPHIC_DOOR_OFFSET, TILE_WIDTH, TILE_HEIGHT, width, height)
			
			--dprintf("player (%d) %d %d", sprite, x, y)
			
			x = x + TILE_WIDTH
			
		end		

	-- Setup exit tiles
	x = 0
	for sprite = 1, MAX_GRAPHIC_EXIT do
		exit_graphic_quads[sprite] = love.graphics.newQuad(x, GRAPHIC_EXIT_OFFSET, TILE_WIDTH, TILE_HEIGHT, width, height)
			
			--dprintf("player (%d) %d %d", sprite, x, y)
			
			x = x + TILE_WIDTH
			
		end
		
	-- Setup special tiles
	x = 0
	for sprite = 1, MAX_GRAPHIC_SPECIAL do
		special_graphic_quads[sprite] = love.graphics.newQuad(x, GRAPHIC_SPECIAL_OFFSET, TILE_WIDTH, TILE_HEIGHT, width, height)
			
			--dprintf("player (%d) %d %d", sprite, x, y)
			
			x = x + TILE_WIDTH
			
		end	

	player_graphic = love.graphics.newImage("players.png")
	width = player_graphic:getWidth()
	height = player_graphic:getHeight()
	
	-- Setup player sprites
	x = 0
	for sprite = 1, MAX_PLAYER_GRAPHIC do
		player_graphic_quads[sprite] = love.graphics.newQuad(x, y, TILE_WIDTH, TILE_HEIGHT, width, height)
		
		--dprintf("player (%d) %d %d", sprite, x, y)
		
		x = x + TILE_WIDTH
		if x >= width then
			x = 0
			y = y + TILE_HEIGHT
		end
	end

	enemy_graphic = love.graphics.newImage("enemies.png")
	width = enemy_graphic:getWidth()
	height = enemy_graphic:getHeight()
	
	-- Setup enemy sprites
	x = 0
	y = 0
	for sprite = 1, MAX_ENEMY_GRAPHIC do
		enemy_graphic_quads[sprite] = love.graphics.newQuad(x, y, TILE_WIDTH, TILE_HEIGHT, width, height)
		
		--dprintf("enemy (%d) %d %d", sprite, x, y)
		
		x = x + TILE_WIDTH
		if x >= width then
			x = 0
			y = y + TILE_HEIGHT
		end
	end

	graphics_table[GRAPHIC_WALLS] = {graphic = tile_graphic, quads = wall_graphic_quads}
	graphics_table[GRAPHIC_DOORS] = {graphic = tile_graphic, quads = door_graphic_quads}
	graphics_table[GRAPHIC_EXITS] = {graphic = tile_graphic, quads = exit_graphic_quads}
	graphics_table[GRAPHIC_SPECIALS] = {graphic = tile_graphic, quads = special_graphic_quads}
	graphics_table[GRAPHIC_ENEMIES] = {graphic = enemy_graphic, quads = enemy_graphic_quads}
	graphics_table[GRAPHIC_PLAYERS] = {graphic = player_graphic, quads = player_graphic_quads}	

end

function draw_level()

	local map_height = #global.map
	local map_width
	local map_value
	local x = 0
	local y = 0
	
	-- Values in the map are range bound.
	for index = 1, map_height do
		map_width = #global.map[index]
		for width = 1, map_width do

			map_value = global.map[index][width]
			
			if map_value < LEVEL_DOOR_RANGE then
				love.graphics.draw(tile_graphic, wall_graphic_quads[map_value], x, y)
			elseif map_value < LEVEL_EXIT_RANGE then
				love.graphics.draw(tile_graphic, door_graphic_quads[map_value - LEVEL_DOOR_RANGE], x, y)				
			elseif map_value < LEVEL_SPECIAL_RANGE then
				love.graphics.draw(tile_graphic, exit_graphic_quads[map_value - LEVEL_EXIT_RANGE], x, y)
			else
				love.graphics.draw(tile_graphic, special_graphic_quads[map_value - LEVEL_SPECIAL_RANGE], x, y)
			end
			
			x = x + TILE_WIDTH
		end
		
		x = 0
		y = y + TILE_HEIGHT
	end

end

function draw_players()

	local quad_offset = 0
	for index = 1, MAX_PLAYERS do
	
		if global.entering_exit==true then
			
			-- Draw disappearing sprites as player is entering the exit.
			if global.entering_exit_player==index then
			
				if global.players[index].disappear_quad==0 then
					-- Set the first disappearing quad for our player.
					global.players[index].disappear_quad = global.players[index].first_disappear_quad
				end

				-- Draw it.
				quad_offset = global.players[index].disappear_quad
				love.graphics.draw(player_graphic, player_graphic_quads[quad_offset], global.players[index].left, global.players[index].top)

				if global.update_frame then

					-- Move to the next frame.
					global.players[index].disappear_quad = global.players[index].disappear_quad + 1
					
					if global.players[index].disappear_quad > global.players[index].first_disappear_quad + MAX_DISAPPEAR_SPRITES then
					
						-- Hit the last disappearing frame.
						global.entering_exit = false
						global.entering_exit_player = 0
						
						-- Mark current player as exited.
						global.players[global.current_player].exited = true
						-- Switch to the next player.
						global.current_player = get_next_player(global.current_player)					
					end					
				end
			end
		end
			
		if global.players[index].active==true and global.players[index].exited==false and
			global.entering_exit_player~=index then
			--dprintf("Drawing player %d %d %d", index, global.players[index].x, global.players[index].y)
			if global.players[index].alive==false then
				love.graphics.draw(tile_graphic, special_graphic_quads[GRAPHIC_SPECIAL_BONES], global.players[index].left, global.players[index].top)
			else
				quad_offset = global.players[index].first_quad + (global.players[index].direction + (global.players[index].frame * MAX_PLAYER_SPRITES))
				love.graphics.draw(player_graphic, player_graphic_quads[quad_offset], global.players[index].left, global.players[index].top)
			end
		end
	end

end

function draw_enemies()

	local enemies = #global.enemies
	for index = 1, enemies do
		local quad_offset = global.enemies[index].first_quad + (global.enemies[index].direction + (global.enemies[index].frame * MAX_PLAYER_SPRITES))
		love.graphics.draw(enemy_graphic, enemy_graphic_quads[quad_offset], global.enemies[index].left, global.enemies[index].top)
	end

end

function draw_weapons()

	local weapons = #global.weapons
	for index = 1, weapons do
		love.graphics.draw(player_graphic, player_graphic_quads[global.weapons[index].quad], global.weapons[index].left, global.weapons[index].top)
	end

end

function draw_players_intro()

	local coords_table=
		{
			{x = TILE_WIDTH, y = TILE_HEIGHT},
			{x = SCREEN_WIDTH - (TILE_WIDTH * 2), y = TILE_HEIGHT},
			{x = TILE_WIDTH, y = SCREEN_HEIGHT - (TILE_HEIGHT * 2)},
			{x = SCREEN_WIDTH - (TILE_WIDTH * 2), y = SCREEN_HEIGHT - (TILE_HEIGHT * 2)},
		}

	local quad_offset = 0
	
	for index = 1, MAX_PLAYERS do
		quad_offset = global.players[index].first_quad + global.intro_direction
		love.graphics.draw(player_graphic, player_graphic_quads[quad_offset], coords_table[index].x, coords_table[index].y)

	end

	if global.update_frame then
		global.intro_direction = global.intro_direction + 1
		if global.intro_direction > MAX_PLAYER_SPRITES then
			global.intro_direction = 1
		end
	end

end

function draw_graphic(index, x, y, index2)

	love.graphics.draw(graphics_table[index2].graphic, graphics_table[index2].quads[index], x, y)

end

