
--[[

levels.lua

(c) bob fossil 2015 - 2016

--]]

require "constants"
require "players"
require "enemies"


LEVEL_WALL_RANGE = 0
LEVEL_DOOR_RANGE = 100
LEVEL_EXIT_RANGE = 200
LEVEL_SPECIAL_RANGE = 300
LEVEL_PLAYER_RANGE = 400
LEVEL_ENEMY_RANGE = 500
-- Range can only go from n to n + 99
LEVEL_RANGE_SIZE = 100

-- Level is maximum of 20 x 13 tiles
MAX_LEVEL_WIDTH = 20
MAX_LEVEL_HEIGHT = 13
MAX_LEVEL_NAME = 40

EXIT_STATE_STABLE = 1
EXIT_STATE_APPEARING = 2
EXIT_STATE_DISAPPEARING = 3

OBJECT_TYPE_KEY = 1
OBJECT_TYPE_POTION = 2
OBJECT_TYPE_FOOD = 3

POTION_KILL_ALL_ENEMIES = 1
POTION_KILL_DEATH = 2
POTION_KILL_ENEMIES = 3

TRAP_ONCE = 1
TRAP_SWITCH = 2
TRAP_PRESSURE = 3

FOOD_ALE = 1
FOOD_MEAT = 2

GENERATOR_OFF = 0
GENERATOR_ON = 1

GENERATOR_MAX_HIT_COUNT = 3

SECTION_NAME = 1
SECTION_DATA = 2
SECTION_EXITS = 3
SECTION_TRAPS = 4
SECTION_GENERATORS = 5
SECTION_HEALTH = 6

MAX_SECTIONS = SECTION_HEALTH

--levels = {}

section_lookup = 
	{
	{name = "[Name]", index = SECTION_NAME},
	{name = "[Data]", index = SECTION_DATA},
	{name = "[Exits]", index = SECTION_EXITS},
	{name = "[Traps]", index = SECTION_TRAPS},
	{name = "[Generators]", index = SECTION_GENERATORS},
	{name = "[Health]", index = SECTION_HEALTH},	
	}
	
--

level_count = 0


global.level_data = {}

-- Level variable name lookup.
level_variable_lookup =
	{
	{id = "CATEGORY_DOOR", value = LEVEL_DOOR_RANGE},
	{id = "CATEGORY_ENEMY", value = LEVEL_ENEMY_RANGE},
	{id = "CATEGORY_EXIT", value = LEVEL_EXIT_RANGE},
	{id = "CATEGORY_SPECIAL", value = LEVEL_SPECIAL_RANGE},
	{id = "CATEGORY_WALL", value = LEVEL_WALL_RANGE},
	{id = "ENEMY_DEATH", value = ENEMY_DEATH},
	{id = "ENEMY_GHOST", value = ENEMY_GHOST},
	{id = "ENEMY_DEMON", value = ENEMY_DEMON},
	{id = "EXIT", value = GRAPHIC_EXIT},
	{id = "EXIT_TO1", value = GRAPHIC_EXIT1},
	{id = "EXIT_TO2", value = GRAPHIC_EXIT2},
	{id = "EXIT_TO3", value = GRAPHIC_EXIT3},
	{id = "EXIT_TO4", value = GRAPHIC_EXIT4},
	{id = "EXIT_TO5", value = GRAPHIC_EXIT5},
	{id = "EXIT_TO6", value = GRAPHIC_EXIT6},
	{id = "EXIT_TO7", value = GRAPHIC_EXIT7},
	{id = "EXIT_TO8", value = GRAPHIC_EXIT8},	
	{id = "INFINITE", value = 0},
	{id = "OFF", value = 0},
	{id = "ON", value = 1},
	{id = "SPECIAL_GROUND", value = GRAPHIC_SPECIAL_GROUND},
	{id = "SPECIAL_AREA1", value = GRAPHIC_SPECIAL_AREA_1},
	{id = "SPECIAL_AREA2", value = GRAPHIC_SPECIAL_AREA_2},
	{id = "SPECIAL_AREA3", value = GRAPHIC_SPECIAL_AREA_3},
	{id = "SPECIAL_AREA4", value = GRAPHIC_SPECIAL_AREA_4},
	{id = "SPECIAL_AREA5", value = GRAPHIC_SPECIAL_AREA_5},
	{id = "SPECIAL_AREA6", value = GRAPHIC_SPECIAL_AREA_6},
	{id = "SPECIAL_AREA7", value = GRAPHIC_SPECIAL_AREA_7},
	{id = "SPECIAL_AREA8", value = GRAPHIC_SPECIAL_AREA_8},
	{id = "SPECIAL_KEY", value = GRAPHIC_SPECIAL_KEY},
--	{id = "SPECIAL_POTION", value = GRAPHIC_SPECIAL_POTION},
	{id = "SPECIAL_POTION_ALL", value =GRAPHIC_SPECIAL_POTION_KILL_ALL},
	{id = "SPECIAL_POTION_DEATH", value =GRAPHIC_SPECIAL_POTION_KILL_DEATH},
	{id = "SPECIAL_POTION_ENEMIES", value =GRAPHIC_SPECIAL_POTION_KILL_ENEMIES},
	{id = "SPECIAL_GENERATOR_OFF", value = GRAPHIC_SPECIAL_GENERATOR_OFF},
	{id = "SPECIAL_GENERATOR_ON", value = GRAPHIC_SPECIAL_GENERATOR_ON},
	{id = "SPECIAL_FORCEFIELD", value = GRAPHIC_SPECIAL_GROUND_FORCEFIELD},
	{id = "SPECIAL_TRAP", value = GRAPHIC_SPECIAL_GROUND_TRAP},
	{id = "SPECIAL_ALE", value = GRAPHIC_SPECIAL_FOOD_ALE},
	{id = "SPECIAL_MEAT", value = GRAPHIC_SPECIAL_FOOD_MEAT},
	{id = "TRAP_ONCE", value = TRAP_ONCE},
	{id = "TRAP_SWITCH", value = TRAP_SWITCH},
	{id = "TRAP_PRESSURE", value = TRAP_PRESSURE},
	{id = "WALL_1", value = GRAPHIC_WALL_1},
	{id = "WALL_2", value = GRAPHIC_WALL_2},
	{id = "WALL_3", value = GRAPHIC_WALL_3},
	{id = "WALL_4", value = GRAPHIC_WALL_4},
	{id = "WALL_5", value = GRAPHIC_WALL_5},
	{id = "WALL_6", value = GRAPHIC_WALL_6},
	{id = "WALL_7", value = GRAPHIC_WALL_7},
	{id = "WALL_8", value = GRAPHIC_WALL_8},
	{id = "WALL_9", value = GRAPHIC_WALL_9},
	{id = "WALL_10", value = GRAPHIC_WALL_10},
	{id = "WALL_11", value = GRAPHIC_WALL_11},
	{id = "WALL_12", value = GRAPHIC_WALL_12},
	{id = "WALL_13", value = GRAPHIC_WALL_13},
	{id = "WALL_14", value = GRAPHIC_WALL_14},
	{id = "WALL_15", value = GRAPHIC_WALL_15},
	{id = "WALL_16", value = GRAPHIC_WALL_16},
	{id = "WALL_DESTRUCTABLE", value = GRAPHIC_WALL_DESTRUCTABLE_1},
	{id = "WALL_MOVEABLE", value = GRAPHIC_WALL_MOVEABLE},
	{id = "WALL_FORCEFIELD", value = GRAPHIC_WALL_FORCEFIELD},	
	}


--[[

Level data:

'Wa-p' Wall pieces
'Wq' Destructable wall piece
'Wt' Moveable wall piece
'Da-o' Door pieces
'Ca-d' Character (a=Warrior, b=Valkyrie, c=Wizard, D=Elf)
'Ea-i' Exit (a=Main Exit, b-i=Teleport Exits 1-8)
'Sa-d' Weapon/Shot (a=Warrior, b=Valkyrie, c=Wizard, D=Elf)
'Ba-b' Baddie (a=Death, b=Ghost)
'_ '   Blank tile
'_a-h' Teleport Exit destination point 1 - 8
'_i'   Forcefield tile
'_t'   Trap tile
'_G' Generator
'Pa-c' Potion (a=Kill everything, b=Kill Death Only, c=Kill Enemies Only)
'k '   Key
'Fa-b' Food (a=ale, b=meat)
--]]

debug_level = 
	{
	name="THE DEBUG ZONE",
	content="WgWkWkWkWkWkWkWpWkWkWkWkWpWkWkWkWkWkWkWm. " ..
		"Wf_b_ _ Bb_ _ Wf_ _ _ _ Wf_ _ _ _ _ _ Wf. " ..
		"Wf_ _ _ _ _ EdWf_ _ _ _ Wf_ _ _ _ _ _ Wf. " ..
		"WhWkWkWkWkWkWkWj_ Ea_ _ WdWkWkWkWkWkWkWn. " ..
		"WfEb_ Sc_ _ _ _ _ _ _ _ _ Sc_ _ _ _ SaWf. " ..
		"Wf_ _ _ _ Ga_ _ _ Cd_ _ _ _ _ _ Bb_ _ Wf. " ..
		"Wf_ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ Wf. " ..
		"Wf_ _ Sd_ _ _ _ _ Cc_ k _ Sd_ _ _ _ SbWf. " ..
		"Wf_ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ Wf. " ..
		"WhWkWkWkWkWkWkWmDjDjDjDjWgWkWkWqWkWkWkWn. " ..
		"Wf_a_ _ _ _ _ Wf_ PaPbPcWf_ _ _ _ _ _ Wf. " ..
		"Wf_ _ _ Ba_ EcWf_ _ _ _ Wf_ _ _ _ _ _ Wf. " ..
		"WdWkWkWkWkWkWkWlWkWkWkWkWlWkWkWkWkWkWkWj.."
	}
	
added_debug_level = false
	
function add_debug_level()

	if added_debug_level==false then
		global.levels = global.levels + 1
		global.level_data[global.levels] = {}
		global.level_data[global.levels].name = debug_level.name
		global.level_data[global.levels].content = debug_level.content
		global.level_data[global.levels].exits = ""
		global.level_data[global.levels].traps = ""
		global.level_data[global.levels].generators = ""
		global.level_data[global.levels].health_data = "999,999,999,999"
		
		global.level = global.levels
		added_debug_level = true
	end
	
	if global.debug_start_level then
		global.level = global.debug_start_level
	end

end

--

function add_level(level_data, level_name, exit_data, trap_data, generator_data, health_data, from_editor)

	if #level_data and #level_name then
		level_count = level_count + 1
		global.level_data[level_count] = {}
		global.level_data[level_count].name = level_name
		global.level_data[level_count].content = level_data
		if exit_data then
			global.level_data[level_count].exits = exit_data
		end
		if trap_data then
			global.level_data[level_count].traps = trap_data
		end
		if generator_data then
			global.level_data[level_count].generators = generator_data
		end
		if health_data then
			global.level_data[level_count].health_data = health_data
		end

		if from_editor then
			level_count = level_count - 1
		else
			dprintf("Added level %d - %s", level_count, global.level_data[level_count].name)
		end
	end

end

function do_variable_expansion(line)

	local start = 1
	local expanded_line = line

	while 1 do
		local dollar, _end = string.find(expanded_line ,"%$", start)
		if dollar then
		
			--dprintf("Found variable line: %s", expanded_line)
		
			local variable = nil
			_, _, variable = string.find(expanded_line, "%$([%w_]+)", start)
			if variable then
			
				local value = -1
				--dprintf("Variable = %s", variable)
				for check = 1, #level_variable_lookup do
					if variable==level_variable_lookup[check].id then
						value = level_variable_lookup[check].value
						break
					end
				end
				
				if value~=-1 then
					local value_string = tostring(value)
					expanded_line = string.gsub(expanded_line, "$".. variable, value_string)
					--dprintf("New line: %s", line)
				end
			end

			start = _end + 1			
			--dprintf("Set start to %d", _end)
		else
			break
		end
		
	end
	
--	if line~=expanded_line then
--		dprintf("Expanded: %s", expanded_line)
--	end

	return expanded_line
end

function load_levels()

	local current_section = 0
	local previous_section = 0
	local level_name = ""
	local level_data = ""
	local exit_data = nil
	local trap_data = nil
	local generator_data = nil
	local health_data = nil
	local line_length = 0
	
	level_count = 0
	global.level_data = {}

	for line in love.filesystem.lines("levels.dat") do

		line_length = #line
		if line_length > 0 then
			local first = string.sub(line, 1, 1)
			-- Check the first character
			if first==";" then
				-- Comment
			elseif first=="[" then
				-- Section header
				for match = 1, MAX_SECTIONS do					
					
					local result = string.find(line, section_lookup[match].name, 1, true)
					if result~=nil then
						previous_section = current_section
						current_section = section_lookup[match].index
						break
					end
				end
			else
				line = do_variable_expansion(line)

				if current_section==SECTION_NAME then
				
					--dprintf("SECTION_NAME")
				
					if previous_section > SECTION_NAME then

						add_level(level_data, level_name, exit_data, trap_data, generator_data, health_data)
					
						level_data = ""
						exit_data = nil
						trap_data = nil
						generator_data = nil
						health_data = nil

					end
					
					if line_length > MAX_LEVEL_NAME then
						level_name = string.sub(line, MAX_LEVEL_NAME)
					else
						level_name = line
					end
					
				elseif current_section==SECTION_DATA then
				
					--dprintf("SECTION_DATA")				
					--dprintf("Level Data: %s", line)
					if previous_section==SECTION_NAME then
						level_data = line
					elseif previous_section==SECTION_DATA then
						level_data = level_data .. line
					end
										
					previous_section = SECTION_DATA
					
				elseif current_section==SECTION_EXITS then

					-- Exit number,{map_x, map_y, exit_type}{map_x1, mapy1, exit_type}....,movement_time)
					if previous_section~=SECTION_EXITS then
						exit_data = string.format("<%s>", line)
					elseif previous_section==SECTION_EXITS then
						exit_data = string.format("%s<%s>", exit_data, line)
					end
				
					previous_section = SECTION_EXITS
					
				elseif current_section==SECTION_TRAPS then

					-- Trap number, trap_type, {map_x, map_y, tile_category, tile_index}{map_x1, mapy1, tile_category, tile_index}....,)
					if previous_section~=SECTION_TRAPS then
						trap_data = string.format("<%s>", line)
					elseif previous_section==SECTION_TRAPS then
						trap_data = string.format("%s<%s>", trap_data, line)
					end
				
					previous_section = SECTION_TRAPS

				elseif current_section==SECTION_GENERATORS then

					-- Generator number, enemy_type, spawn time, maximum_spawn
					if previous_section~=SECTION_GENERATORS then
						generator_data = string.format("<%s>", line)
					elseif previous_section==SECTION_GENERATORS then
						generator_data = string.format("%s<%s>", generator_data, line)
					end
				
					previous_section = SECTION_GENERATORS
					
				elseif current_section==SECTION_HEALTH then
				
					--dprintf("SECTION_HEALTH")				
					--dprintf("Health data: %s", line)
					
					health_data = line
									
					previous_section = SECTION_HEALTH
				end
			end
		end
	end
	
	-- Add the last level loaded.
	if previous_section > SECTION_NAME then

		add_level(level_data, level_name, exit_data, trap_data, generator_data, health_data)

		level_data = ""
		exit_data = nil
		trap_data = nil
		generator_data = nil
		health_data = nil
	
	end
	
	global.levels = level_count

end

function create_map_from_level_data(level_number, from_editor)

	local map = {}
	local content_length = #global.level_data[level_number].content
	local map_x = 1
	local map_y = 1
	local id = 0
	local param = 0
	local value = 0

	-- Convert tile string to array of values.
	map[map_y]={}
	
	-- Set health values.
	_, _, h1, h2, h3, h4 = string.find(global.level_data[level_number].health_data, "(%d+),(%d+),(%d+),(%d+)")
	global.level_data[level_number].health = {}
	global.level_data[level_number].health[PLAYER_WARRIOR] = tonumber(h1)
	global.level_data[level_number].health[PLAYER_VALKYRIE] = tonumber(h2)
	global.level_data[level_number].health[PLAYER_WIZARD] = tonumber(h3)
	global.level_data[level_number].health[PLAYER_ELF] = tonumber(h4)
	
	for index = 1, content_length, 2 do
		-- Get value from tile string.
		value = 0
		id = string.byte(global.level_data[level_number].content, index)
		param = string.byte(global.level_data[level_number].content, index + 1)
		
		if id==46 then
			-- . Full stop
			-- If param is . then we've hit the end.
			if param~=46 then
				map_x = 1
				map_y = map_y + 1
				map[map_y]={}
			end
		elseif id==66 then
			-- B Bad-guy
			-- param is lowercase letter a-z
			if not from_editor then
				add_enemy(param - 96, (map_x - 1) * TILE_WIDTH, (map_y - 1) * TILE_HEIGHT, 0)
			end
			
			if from_editor then
				value = (param - 96) + LEVEL_ENEMY_RANGE
			else
				-- Put enemy on safe ground.
				value = GRAPHIC_SPECIAL_GROUND + LEVEL_SPECIAL_RANGE
			end
			
		elseif id==67 then
			-- C Character
			-- param is lowercase letter a-d
			local health = global.level_data[level_number].health[param - 96]
			add_player(param - 96, (map_x - 1) * TILE_WIDTH, (map_y - 1) * TILE_HEIGHT, health)

			-- Put player on safe ground.
			value = GRAPHIC_SPECIAL_GROUND + LEVEL_SPECIAL_RANGE
			
		elseif id==68 then
			-- D Door
			-- param is lowercase letter a-z
			value = (param - 96) + LEVEL_DOOR_RANGE
		elseif id==69 then
			-- E Exit
			-- param is 
			value = (param - 96) + LEVEL_EXIT_RANGE
			if not from_editor then
				local exits = #global.exits
				global.exits[exits + 1] = {left = (map_x - 1) * TILE_WIDTH, top = (map_y - 1) * TILE_HEIGHT,
					right = (map_x * TILE_WIDTH), bottom = (map_y * TILE_HEIGHT), _type = param - 97, state = EXIT_STATE_STABLE}
			end
		elseif id==70 then
			-- F Food
			-- param is lowercase letter a-z
			value = GRAPHIC_SPECIAL_FOOD_ALE + LEVEL_SPECIAL_RANGE + (param - 97)
			if not from_editor then
				local objects = #global.level_objects
				global.level_objects[objects + 1] = {left = (map_x - 1) * TILE_WIDTH, top = (map_y - 1) * TILE_HEIGHT,
					right = (map_x * TILE_WIDTH), bottom = (map_y * TILE_HEIGHT), collected = false, _type = OBJECT_TYPE_FOOD, food_type = param - 96}
			end			
		elseif id==80 then
			-- P Potion
			-- param is lowercase letter a-z
			value = GRAPHIC_SPECIAL_POTION + LEVEL_SPECIAL_RANGE
			if not from_editor then
				local objects = #global.level_objects
				global.level_objects[objects + 1] = {left = (map_x - 1) * TILE_WIDTH, top = (map_y - 1) * TILE_HEIGHT,
					right = (map_x * TILE_WIDTH), bottom = (map_y * TILE_HEIGHT), collected = false, _type = OBJECT_TYPE_POTION, potion_type = param - 96}
			end
		elseif id==87 then
			-- W Wall
			-- param is lowercase letter a-z
			value = (param - 96) + LEVEL_WALL_RANGE				
		elseif id==95 then
			-- _ Underscore
			if param==32 then	-- ' '
				value = GRAPHIC_SPECIAL_GROUND + LEVEL_SPECIAL_RANGE
			elseif param==71 then	-- 'G'
				value = GRAPHIC_SPECIAL_GENERATOR_ON + LEVEL_SPECIAL_RANGE
				
				if not from_editor then
					local generators = #global.generators
					global.generators[generators + 1] = {left = (map_x - 1) * TILE_WIDTH, top = (map_y - 1) * TILE_HEIGHT,
						right = (map_x * TILE_WIDTH), bottom = (map_y * TILE_HEIGHT)}
				end				
			elseif param==105 then	-- 'i'
				value = GRAPHIC_SPECIAL_GROUND_FORCEFIELD + LEVEL_SPECIAL_RANGE
			elseif param==116 then	-- 't'
				value = GRAPHIC_SPECIAL_GROUND_TRAP + LEVEL_SPECIAL_RANGE
				
				if not from_editor then
					local traps = #global.traps
					global.traps[traps + 1] = {left = (map_x - 1) * TILE_WIDTH, top = (map_y - 1) * TILE_HEIGHT,
						right = (map_x * TILE_WIDTH), bottom = (map_y * TILE_HEIGHT)}
				end
			else
				value = (param - 96) + GRAPHIC_SPECIAL_GROUND + LEVEL_SPECIAL_RANGE
				
				if not from_editor then
					local areas = #global.areas
					global.areas[areas + 1] = {left = (map_x - 1) * TILE_WIDTH, top = (map_y - 1) * TILE_HEIGHT,
						right = (map_x * TILE_WIDTH), bottom = (map_y * TILE_HEIGHT), id = param - 96}	
				end
			end
		elseif id==107 then
			-- k Key
			value = GRAPHIC_SPECIAL_KEY + LEVEL_SPECIAL_RANGE
			if not from_editor then
				local objects = #global.level_objects
				global.level_objects[objects + 1] = {left = (map_x - 1) * TILE_WIDTH, top = (map_y - 1) * TILE_HEIGHT,
					right = (map_x * TILE_WIDTH), bottom = (map_y * TILE_HEIGHT), collected = false, _type = OBJECT_TYPE_KEY}
			end
		end
		
		if value~=0 then
			map[map_y][map_x] = value
			map_x = map_x + 1
		end		
	end

	return map

end

function init_level(level_number)

	reset_players()	
	
	global.map = {}
	global.exits = {}
	global.areas = {}
	global.enemies = {}
	global.weapons = {}
	global.generators = {}
	global.traps = {}
	global.level_objects = {}
	
	global.map = create_map_from_level_data(level_number, false)
	
	--dump_map()
	
	if #global.exits and global.level_data[level_number].exits then
	
		--dprintf("Exits = %s", global.level_data[level_number].exits)
		for item in string.gmatch(global.level_data[level_number].exits, "<[0-9,{}]+>") do

			local exit_line = string.sub(item, 2, -2)
			--dprintf("%s", exit_line)
		     
			-- Add in extra properties for exits.
			local _index, data, timeout = string.match(exit_line, "(%d+),(.+),(%d+)")
			local index = tonumber(_index)
			--dprintf("Exit %d, Timeout %d %d %d", index, timeout, #global.exits[1], #global.exits)
			
			global.exits[index].timeout = tonumber(timeout)
			global.exits[index].movement = {}
			global.exits[index].current_time = global.current_time
			global.exits[index].movement_index = 1
			
			for item in string.gmatch(data, "{%d+,%d+,%d+}") do
			     local mx, my, exit_tile = string.match(item, "{(%d+),(%d+),(%d+)}")
			     global.exits[index].movement[#global.exits[index].movement + 1] = {map_x = tonumber(mx),
				map_y = tonumber(my), tile = tonumber(exit_tile)}

			     --dprintf("map_x %d map_y %d exit_tile %d", mx, my, exit_tile)
			end
		end		
	end

	if #global.traps and global.level_data[level_number].traps then
		-- Add in extra properties for traps.
		for item in string.gmatch(global.level_data[level_number].traps, "<[0-9,{}]+>") do
		     
		     local trap_line = string.sub(item, 2, -2)
		     --dprintf("%s", trap_line)

			-- Add properties for traps.
			local _index, __type, data = string.match(trap_line, "(%d+),(%d+),(.+)")
			local index = tonumber(_index)
			local _type = tonumber(__type)
			
			--dprintf("index: %d type: %d data %s", index, _type, data)
			
			global.traps[index].tiles = {}
			global.traps[index].activated = false
			global.traps[index].trap_type = _type
			global.traps[index].player = 0
			
			for item in string.gmatch(data, "{%d+,%d+,%d+,%d+}") do
			     local mx, my, category, tile = string.match(item, "{(%d+),(%d+),(%d+),(%d+)}")
			     global.traps[index].tiles[#global.traps[index].tiles+ 1] = {map_x = tonumber(mx),
				map_y = tonumber(my), tile_category = tonumber(category), tile_index = tonumber(tile), on = false}

			    --dprintf("map_x %d map_y %d tile category %d tile %d", mx, my, category, tile)
			end
		end
	end
	
	if #global.generators and global.level_data[level_number].generators then
		-- Add in extra properties for generators.
		for item in string.gmatch(global.level_data[level_number].generators, "<[0-9,{}]+>") do
		     
		     local generator_line = string.sub(item, 2, -2)
		     --dprintf("%s", generator_line)
		     
			local _index, _enemy_type, _spawn_time, _spawn_count, _state = string.match(generator_line, "(%d+),(%d+),(%d+),(%d+),(%d+)")
			local index = tonumber(_index)
			local generator = global.generators[index] 
			generator.enemy_type = tonumber(_enemy_type)
 			generator.spawn_time = tonumber(_spawn_time)
			generator.spawn_count = tonumber(_spawn_count)
			generator.state = tonumber(_state)
			
			local map_x = (generator.left / TILE_WIDTH) + 1
			local map_y = (generator.top / TILE_HEIGHT) + 1
			if generator.state==GENERATOR_OFF then
				global.map[map_y][map_x] =
					LEVEL_SPECIAL_RANGE + GRAPHIC_SPECIAL_GENERATOR_OFF
			elseif generator.state==GENERATOR_ON then
				global.map[map_y][map_x] =
					LEVEL_SPECIAL_RANGE + GRAPHIC_SPECIAL_GENERATOR_ON
			end
			
			-- Runtime stuff.
			generator.spawned = 0
			generator.current_time = global.current_time
			generator.next_slot = 0
			generator.hit_count = 0
			generator.slot_table = {false, false, false, false, false, false, false, false}
		end
	end
	
	global.current_player = get_first_player()
	global.entering_exit = false
	global.entering_exit_player = 0
	
	-- Set firing_time to the current time to stop the space keypress from message
	-- confirmation being treated as a spurious fire press.
	global.players[global.current_player].firing_time = global.current_time

end

function get_level_name()

	return global.level_data[global.level].name

end

function next_level()

	if global.state==STATE_EDITOR_TEST then
		global.state = STATE_EDITOR
	else
		-- Move on to the next level.
		global.level = global.level + 1
		if global.level > global.levels then
			-- Wrap around.
			global.level = 1
		end
		
		dprintf("Level is: %d", global.level)
		
		init_level(global.level)
		global.state = STATE_INGAME
	end

end


function restart_level()

	if global.state==STATE_EDITOR_TEST then
		global.state = STATE_EDITOR
	else
		-- Reintialise the current level.
		init_level(global.level)
	end

end

function quit_level()

	if global.state==STATE_EDITOR_TEST then
		global.state = STATE_EDITOR
	else
		-- Quit back to the intro screen.
		init_titlescreen()
	end

end

--

function map_to_string(map, output)

	local map_string = ""
	if output then
		map_string = "[Name]\r\nMADE IN THE LEVEL EDITOR\r\n[Data]\r\n"
	end
		
	local special_tile_lookup =
		{
		{tile = "_ "},	-- GRAPHIC_SPECIAL_GROUND
		{tile = "_a"},	-- GRAPHIC_SPECIAL_AREA_1
		{tile = "_b"},	-- GRAPHIC_SPECIAL_AREA_2
		{tile = "_c"},	-- GRAPHIC_SPECIAL_AREA_3
		{tile = "_d"},	-- GRAPHIC_SPECIAL_AREA_4
		{tile = "_e"},	-- GRAPHIC_SPECIAL_AREA_5
		{tile = "_f"},	-- GRAPHIC_SPECIAL_AREA_6
		{tile = "_g"},	-- GRAPHIC_SPECIAL_AREA_7
		{tile = "_h"},	-- GRAPHIC_SPECIAL_AREA_8
		{tile = "k "},	-- GRAPHIC_SPECIAL_KEY
		{tile = "Pa"},	-- GRAPHIC_SPECIAL_POTION
		{tile = "!!"},	-- GRAPHIC_SPECIAL_BONES
		{tile = "!!"},	-- GRAPHIC_SPECIAL_GENERATOR_DAMAGED
		{tile = "!!"},	-- GRAPHIC_SPECIAL_GENERATOR_OFF
		{tile = "_G"},	-- GRAPHIC_SPECIAL_GENERATOR_ON
		{tile = "_i"},	-- GRAPHIC_SPECIAL_GROUND_FORCEFIELD
		{tile = "_t"},	-- GRAPHIC_SPECIAL_GROUND_TRAP
		{tile = "Fa"},	-- GRAPHIC_SPECIAL_FOOD_ALE
		{tile = "Fb"},	-- GRAPHIC_SPECIAL_FOOD_MEAT
		}

	local map_height = #map

	for index = 1, map_height do
		local tile = ""
		local map_width = #map[index]
		for width = 1, map_width do
			local value = map[index][width]

			if value >= LEVEL_ENEMY_RANGE then
				tile = "B"
				value = (value - LEVEL_ENEMY_RANGE) + 96
			elseif value >= LEVEL_PLAYER_RANGE then
				tile = "C"
				value = (value - LEVEL_PLAYER_RANGE) + 96
			elseif value >= LEVEL_SPECIAL_RANGE then
				tile = special_tile_lookup[value - LEVEL_SPECIAL_RANGE].tile
				value = 0
			elseif value >= LEVEL_EXIT_RANGE then
				tile = "E"
				value = (value - LEVEL_EXIT_RANGE) + 96
			elseif value >= LEVEL_DOOR_RANGE then
				tile = "D"
				value = (value - LEVEL_DOOR_RANGE) + 96			
			elseif value >= LEVEL_WALL_RANGE then	
				tile = "W"
				value = (value - LEVEL_WALL_RANGE) + 96
			else
				tile = "?"
			end
			
			if value==0 then
				map_string = map_string .. tile
			else
				map_string = map_string .. string.format("%s%c", tile, value)
			end
		end
		
		if index==map_height then
			-- Last line.
			map_string = map_string .. ".."
			if output then
				map_string = map_string .. "\r\n\r\n"
			end
		else
			map_string = map_string .. ". "
			if output then
				map_string = map_string .. "\r\n"
			end
		end
	end
	
	return map_string
end
