
--[[

Puzzlet

main.lua

(c) bob fossil 2015-2016

--]]

--[[

To do:

- Debug:
-
- Weapons:
-
- Exits:
-
- Enemies:
- 	Generators
-		- Ghosts
-	Death
-
- Levels:
-	- More of them!
-
- Other:
-	Redefinable controls
-	Editor improvements
- Tidy / comment / refactor :)

-]]

global = {}
global.debug_mode = false
global.current_time = 0
global.last_frame_time = 0
global.last_health_time = 0
global.state = 0
global.level = 0
global.levels = 0
global.current_player = 0
global.update_frame = false
global.levels = {}
global.players={}
global.weapons={}
global.generators={}
global.map = {}
global.intro_colour_index = 1
global.intro_direction = 1

require "audio"
require "constants"
require "actions"
require "editor"
require "font"
require "game"
require "levels"
require "graphics"
require "instructions"
require "message"
require "titlescreen"

require "my_debug"

WINDOW_TITLE = "Puzzlet"
SAVE_FOLDER = "puzzlet"
WINDOW_VERSION = "WIP 0.02"

state_functions = {}

function init_controls()
	-- !!! Not implemented
end



function init_game()

	global.level = 1

	if global.debug_mode==true then
		add_debug_level()
	end
	
	init_level(global.level)

	-- Switch to game state.
	global.state = STATE_INGAME
end

function love.load()
	
	-- Set up the main window.
	love.graphics.setBackgroundColor(0, 0, 0)
	
	love.window.setMode(SCREEN_WIDTH, SCREEN_HEIGHT, {fullscreen=false, vsync=true, msaa=0})
	love.window.setTitle(WINDOW_TITLE .. " " .. WINDOW_VERSION)
	
	love.filesystem.setIdentity(SAVE_FOLDER)
	
	-- Set up our state function jump table.
	-- Lets us do a switch style selection of state functions rather than a load of
	-- if...elseif statements.
	state_functions[STATE_TITLESCREEN] = {input = handle_titlescreen_keypress,
		draw = draw_titlescreen}
	state_functions[STATE_INGAME] = {input = handle_game_keypress, draw = draw_game,
		update = update_game}
	state_functions[STATE_INMESSAGE] = {input = handle_message_keypress,
		draw = draw_message}
	state_functions[STATE_INSTRUCTIONS] = {input = handle_instruction_keypress,
		draw = draw_instructions}
	state_functions[STATE_EDITOR] = {input = handle_editor_keypress, draw = draw_editor}
	state_functions[STATE_EDITOR_TEST] = {input = handle_game_keypress, draw = draw_game,
		update = update_game}	
	state_functions[STATE_INSPEECH] = {}
	
	init_font()
	init_graphics()
	init_audio()
	
	load_levels()

	reset_players()
	
	init_titlescreen()

end

function love.update(dt)

	global.current_time = global.current_time + dt * 1000
	-- See if enought time has passed for an animation frame update.
	
	if global.current_time - global.last_frame_time > 250 then
		global.update_frame = true
		global.last_frame_time = global.current_time
	end
	
	-- Call the appropriate update function.
	if state_functions[global.state].update then
		state_functions[global.state].update()
	end
end

function love.keypressed(key, scancode, is_repeat)

	-- Call the appropriate key handling function.
	if state_functions[global.state].input then
		state_functions[global.state].input(key)
	end

end

function love.draw()

	-- Call the appropriate drawing function.
	if state_functions[global.state].draw then
		state_functions[global.state].draw()
	end

	if global.update_frame then
		global.update_frame = false
	end
	
end

