
--[[

Puzzlet

message.lua

(c) bob fossil 2016

--]]

MESSAGE_MENU = 1
MESSAGE_NEXT_LEVEL = 2
MESSAGE_RESTART_QUIT_LEVEL = 3
MESSAGE_RESTART_ALL_DEAD = 4
MESSAGE_RESTART_SOME_DEAD = 5

message_table=
	{
	{text = {"              "}, options={" START GAME ", "  CONTROLS  ", "INSTRUCTIONS", "   EDITOR   ", "    EXIT    "}, initial = 1, custom_colour={r = 128, b = 0, g = 0}, edge = {r = 255, g = 255, b = 255}, vertical = true},
	{text = {"     LEVEL CLEARED!     ", "", "PRESS SPACE TO CONTINUE."}, custom_colour={ r =0, g = 128, b = 128}},--colour = PLAYER_VALKYRIE},
	{text = {"RESTART OR QUIT THIS LEVEL?"}, options={"RESTART", "QUIT"}, initial = 1, colour = PLAYER_ELF},
	{text = {"NO ONE MADE IT OUT ALIVE!", "", " PRESS SPACE TO RESTART. "}, colour = PLAYER_WARRIOR},
	{text = {"NOT EVERYBODY EXITED ALIVE!", "", "  PRESS SPACE TO RESTART.  "}, colour = PLAYER_WARRIOR},	
	}
	
current_message_index = 0
current_message_option = 0
previous_state = 0
current_functions={}

function set_message(message_index, message_functions)

	previous_state = global.state
	global.state = STATE_INMESSAGE
	current_message_index = message_index
	current_message_option = 0
	current_functions = {}
	if message_functions then
		for copy = 1, #message_functions do
			current_functions[copy] = message_functions[copy]			
		end
	end
	
	current_message_option = 1
	if message_table[current_message_index].initial then
		-- Select an initial option.
		current_message_option = message_table[current_message_index].initial
	end
	
end

function handle_message_keypress(key)

	if key=="space" or key=="return" then
	
		global.state = previous_state
		if #current_functions > 0 then
			-- Do the option we selected.			
			current_functions[current_message_option]()
		end
	
	elseif key=="left" then
		
		if message_table[current_message_index].options then
		
			current_message_option = current_message_option - 1
			if current_message_option < 1 then
				current_message_option = #message_table[current_message_index].options
			end
		
		end
		
	elseif key=="right" then
		
		if message_table[current_message_index].options then
		
			current_message_option = current_message_option + 1
			if current_message_option > #message_table[current_message_index].options then
				current_message_option = 1
			end
		
		end
		
	elseif key=="up" then
		
		if message_table[current_message_index].options and
			message_table[current_message_index].vertical then
		
			current_message_option = current_message_option - 1
			if current_message_option < 1 then
				current_message_option = #message_table[current_message_index].options
			end
		
		end
		
	elseif key=="down" then
		
		if message_table[current_message_index].options and
			message_table[current_message_index].vertical then
		
			current_message_option = current_message_option + 1
			if current_message_option > #message_table[current_message_index].options then
				current_message_option = 1
			end
		
		end		
	
	elseif key=="escape" then
		
		global.state = previous_state
	end
	

end
	
function draw_message()

	if previous_state==STATE_INGAME or previous_state==STATE_EDITOR_TEST then

		-- Dim the screen
		love.graphics.setColor(255, 255, 255, 128)

		draw_level()
		draw_players()
	end

	local r
	local g
	local b
	if message_table[current_message_index].custom_colour then

		r = message_table[current_message_index].custom_colour.r
		g = message_table[current_message_index].custom_colour.g
		b = message_table[current_message_index].custom_colour.b
	else
		r = global.players[message_table[current_message_index].colour].r
		g = global.players[message_table[current_message_index].colour].g
		b = global.players[message_table[current_message_index].colour].b
	end
	
	love.graphics.setColor(r, g, b, 255)

	local message_lines = #message_table[current_message_index].text
	local options = 0
	if message_table[current_message_index].options then
		options = #message_table[current_message_index].options
	end
	
	-- Work out dimensions of the message box.
	local lines = message_lines + 2
	local length = 0
	for line = 1, message_lines do
		if #message_table[current_message_index].text[line] > length then
			length = #message_table[current_message_index].text[line]
		end
	end
	
	if message_table[current_message_index].vertical then
		lines = lines + options
	end

	length = length + 2
	
	if options > 0 then
		lines = lines + 3
	end

	-- Draw background.
	local width = length * FONT_CHARACTER_WIDTH
	local height = lines * FONT_CHARACTER_HEIGHT
	local x = (SCREEN_WIDTH - width) / 2
	local y = (SCREEN_HEIGHT - height) / 2
	love.graphics.rectangle("fill", x, y, width, height)
	
	if message_table[current_message_index].edge then
		-- Edge colour specified so draw a border in that colour.
		love.graphics.setColor(message_table[current_message_index].edge.r,
			message_table[current_message_index].edge.g,
			message_table[current_message_index].edge.b, 255)
			love.graphics.rectangle("line", x, y, width, height)
	end

	-- Draw lines of text.
	for line = 1, message_lines do
		draw_text(x + FONT_CHARACTER_WIDTH + 1, y + FONT_CHARACTER_HEIGHT + 1, message_table[current_message_index].text[line], 0, 0, 0)
		draw_text(x + FONT_CHARACTER_WIDTH, y + FONT_CHARACTER_HEIGHT, message_table[current_message_index].text[line], 255, 255, 255)
		y = y + FONT_CHARACTER_HEIGHT
	end
	
	if options > 0 then
		-- Draw options.
		if message_table[current_message_index].vertical then
		
			y = y + (FONT_CHARACTER_HEIGHT * 2)
			for option = 1, options do
				local offset_x = ((width - (#message_table[current_message_index].options[option] * FONT_CHARACTER_WIDTH)) / 2)
				
				-- Is this the selected item?
				if option==current_message_option then
					love.graphics.setColor(r / 2, g / 2, b / 2, 255)
					love.graphics.rectangle("fill", x + FONT_CHARACTER_WIDTH, y, width - (FONT_CHARACTER_WIDTH * 2), FONT_CHARACTER_HEIGHT)
				end
				
				draw_text(x + offset_x + 1, y + 1, message_table[current_message_index].options[option], 0, 0, 0)
				draw_text(x + offset_x, y, message_table[current_message_index].options[option], 255, 255, 255)	
				y = y + FONT_CHARACTER_HEIGHT

			end
		else
			local options_width = width / options
			y = y + (FONT_CHARACTER_HEIGHT * 2)
			for option = 1, options do
				local offset_x = (x + options_width - x) / 2
				offset_x = offset_x - ((#message_table[current_message_index].options[option] * FONT_CHARACTER_WIDTH) / 2)
			
				-- Is this the selected item?
				if option==current_message_option then
					love.graphics.setColor(r / 2, g / 2, b / 2, 255)
					love.graphics.rectangle("fill", x + FONT_CHARACTER_WIDTH, y, options_width - (FONT_CHARACTER_WIDTH * 2), FONT_CHARACTER_HEIGHT)
				end

				draw_text(x + offset_x + 1, y + 1, message_table[current_message_index].options[option], 0, 0, 0)
				draw_text(x + offset_x, y, message_table[current_message_index].options[option], 255, 255, 255)
				x = x + options_width
			end
		end
	end

end

