
--[[

Puzzlet

players.lua

(c) bob fossil 2015-2016

--]]

require "constants"

global.players =
	{
	{name = "WARRIOR", r = 255, g = 0, b = 0},
	{name = "VALKYRIE", r = 0, g = 255, b = 255},
	{name = "WIZARD", r = 255, g = 255, b = 0},
	{name = "ELF", r = 0, g = 255, b = 0},	
	}
	
PLAYERS_ALL_EXITED = 1
PLAYERS_ALL_DEAD = 2
PLAYERS_SOME_EXITED = 3

PLAYER_HEALTH_TIME_PERIOD = 1000
PLAYER_LOW_HEALTH = 20

function reset_players()

	for add = 1, MAX_PLAYERS do
	
		global.players[add].active = false
		global.players[add].alive = false
		global.players[add].exited = false
		global.players[add].left = 0
		global.players[add].top = 0
		global.players[add].right = 0
		global.players[add].bottom = 0		
		global.players[add].keys = 0
		global.players[add].health = 0
		global.players[add].status = {running = false, shooting = false, custom_action = false}
		global.players[add].frame = 0
	end

	global.players[PLAYER_WARRIOR].first_quad = GRAPHIC_WARRIOR_UP_1 - 1
	global.players[PLAYER_WARRIOR].disappear_quad = 0
	global.players[PLAYER_WARRIOR].first_disappear_quad = GRAPHIC_WARRIOR_DISAPPEAR_1
	global.players[PLAYER_WARRIOR].first_weapon_quad = GRAPHIC_WARRIOR_WEAPON_UP - 1
	global.players[PLAYER_WARRIOR].direction = DOWN
	
	global.players[PLAYER_VALKYRIE].first_quad = GRAPHIC_VALKYRIE_UP_1 - 1
	global.players[PLAYER_VALKYRIE].disappear_quad = 0
	global.players[PLAYER_VALKYRIE].first_disappear_quad = GRAPHIC_VALKYRIE_DISAPPEAR_1
	global.players[PLAYER_VALKYRIE].first_weapon_quad = GRAPHIC_VALKYRIE_WEAPON_UP - 1
	global.players[PLAYER_VALKYRIE].direction = DOWN

	global.players[PLAYER_WIZARD].first_quad = GRAPHIC_WIZARD_UP_1 - 1
	global.players[PLAYER_WIZARD].disappear_quad = 0
	global.players[PLAYER_WIZARD].first_disappear_quad = GRAPHIC_WIZARD_DISAPPEAR_1
	global.players[PLAYER_WIZARD].first_weapon_quad = GRAPHIC_WIZARD_WEAPON_UP - 1
	global.players[PLAYER_WIZARD].direction = DOWN

	global.players[PLAYER_ELF].first_quad = GRAPHIC_ELF_UP_1 - 1
	global.players[PLAYER_ELF].disappear_quad = 0
	global.players[PLAYER_ELF].first_disappear_quad = GRAPHIC_ELF_DISAPPEAR_1
	global.players[PLAYER_ELF].first_weapon_quad = GRAPHIC_ELF_WEAPON_UP - 1
	global.players[PLAYER_ELF].direction = DOWN
		
end

function add_player(index, x, y, health)

	if global.debug_mode then
		dprintf("Added player %s at %d %d, health %d", global.players[index].name, x, y, health)
	end

	global.players[index].active = true
	global.players[index].alive = true
	global.players[index].left = x
	global.players[index].top = y
	global.players[index].right = x + TILE_WIDTH
	global.players[index].bottom = y + TILE_HEIGHT
	global.players[index].health = health
	global.players[index].status = {running = false, shooting = false, custom_action = false}

end

function get_first_player()

	for check = 1, MAX_PLAYERS do

		if global.players[check].active == true and global.players[check].alive == true then
			return check
		end
	end

end

-- Returns true if we can switch to the specified player.
function can_switch_to_player(player_index)

	if global.players[player_index].active == true and
		global.players[player_index].alive == true and 
		global.players[player_index].exited == false then
			return true
	end
	
	return false

end

function get_next_player(current)

	local found = false
	local start = current
	local count = 1
	while 1 do
		
		current = current + 1
		if current > MAX_PLAYERS then
			current = 1
		end
		
		if current==start then
			return current
		end
	
		if global.players[current].active == true and global.players[current].alive == true 
			and global.players[current].exited == false then
			return current
		end
	end

	return 0
end

function check_players()

	local dead = 0
	local exited = 0
	local active = 0

	-- Check the active players for this level.
	for check = 1, MAX_PLAYERS do

		if global.players[check].active == true then

			active = active + 1 
			if global.players[check].alive == false then
				dead = dead + 1
			elseif global.players[check].exited == true then
				exited = exited + 1
			end
		end
	end
	
	--dprintf("check_players: dead %d exitted %d active %d", dead, exited, active)
	
	if dead==active then
		-- All active players are dead.
		return PLAYERS_ALL_DEAD
	end
	
	if exited==active then
		-- All active players have exited the level.
		return PLAYERS_ALL_EXITED
	end
	
	if exited + dead==active then
		return PLAYERS_SOME_EXITED
	end
	
	return 0
		

end

function set_player_position(player_index, new_x, new_y, new_direction)

	--dprintf("nd %d d %d", new_direction, global.players[player_index].direction)

	if new_direction~=global.players[player_index].direction then
		
		global.players[player_index].frame = 0
		global.players[player_index].direction = new_direction
		
	else
		if global.update_frame==true then
			global.players[player_index].frame = global.players[player_index].frame + 1
			if global.players[player_index].frame > MAX_FRAMES then
			
				global.players[player_index].frame = 0
			end			
		end
		
	end
	
	--dprintf("f %d - %d", global.players[player_index].frame, global.players[player_index].first_quad)
	
	global.players[player_index].left = new_x
	global.players[player_index].top = new_y
	global.players[player_index].right = new_x + TILE_WIDTH
	global.players[player_index].bottom = new_y + TILE_HEIGHT
end