
--[[

Puzzlet

weapons.lua

(c) bob fossil 2016

--]]

WEAPON_TIME_INTERVAL = 250
WEAPON_SCALE_X = 4
WEAPON_SCALE_Y = 4

weapon_direction_offsets =
	{
	{x = 0, y = -TILE_HEIGHT, mirror = DOWN, hot_x = TILE_WIDTH / 2, hot_y = 0},			-- UP 
	{x = TILE_WIDTH, y = -TILE_HEIGHT, mirror = DOWN_LEFT, hot_x = TILE_WIDTH - 1, hot_y = 0},		-- UP_RIGHT
	{x = TILE_WIDTH, y = 0, mirror = LEFT, hot_x = TILE_WIDTH - 1, hot_y = TILE_HEIGHT / 2},		-- RIGHT
	{x = TILE_WIDTH, y = TILE_HEIGHT, mirror = UP_LEFT, hot_x = TILE_WIDTH - 1, hot_y = TILE_HEIGHT - 1},	-- DOWN_RIGHT
	{x = 0, y = TILE_HEIGHT, mirror = UP, hot_x = TILE_WIDTH / 2, hot_y = TILE_HEIGHT - 1},		-- DOWN
	{x = -TILE_WIDTH, y = TILE_HEIGHT, mirror = UP_RIGHT, hot_x = 0, hot_y = TILE_HEIGHT - 1},		-- DOWN_LEFT
	{x = -TILE_WIDTH, y = 0, mirror = RIGHT, hot_x = 0, hot_y = TILE_HEIGHT / 2},			-- LEFT
	{x = -TILE_WIDTH, y = -TILE_HEIGHT, mirror = DOWN_RIGHT, hot_x = 0, hot_y = 0},			-- UP_LEFT
	}
	
function hit_generator(map_x, map_y, player)

	local hit_x = (map_x - 1) * TILE_WIDTH
	local hit_y = (map_y - 1) * TILE_HEIGHT
	
	for generator_index = 1, #global.generators do

		local gx = global.generators[generator_index].left
		local gy = global.generators[generator_index].top
		
		if gx==hit_x and gy==hit_y then

			local generator = global.generators[generator_index]
			if player==PLAYER_ELF then

				if generator.state==GENERATOR_ON then
					--dprintf("Elf arrow switched generator off!")
					generator.state = GENERATOR_OFF
					
					global.map[map_y][map_x] = LEVEL_SPECIAL_RANGE +
						GRAPHIC_SPECIAL_GENERATOR_OFF
						
				elseif generator.state==GENERATOR_OFF then
					--dprintf("Elf arrow switched generator on!")
					generator.state = GENERATOR_ON
					generator.current_time = global.current_time
					
					global.map[map_y][map_x] = LEVEL_SPECIAL_RANGE +
						GRAPHIC_SPECIAL_GENERATOR_ON
				end
				
			else
				--dprintf("Shot damaged generator")
				generator.hit_count = generator.hit_count + 1
				if generator.hit_count > GENERATOR_MAX_HIT_COUNT then
					-- Remove this generator as it's been hit.
					global.map[map_y][map_x] = LEVEL_SPECIAL_RANGE +
						GRAPHIC_SPECIAL_GROUND
					table.remove(global.generators, generator_index)
				else
					-- Show the generator has been damaged.
					global.map[map_y][map_x] = LEVEL_SPECIAL_RANGE +
						GRAPHIC_SPECIAL_GENERATOR_DAMAGED
				end
			end
			
			break
		end
	end

end

function hit_trap(map_x, map_y)

	local hit_x = (map_x - 1) * TILE_WIDTH
	local hit_y = (map_y - 1) * TILE_HEIGHT
	
	for trap = 1, #global.traps do

		local tx = global.traps[trap].left
		local ty = global.traps[trap].top
		
		if tx==hit_x and ty==hit_y then

			-- Pressure traps can't be activated by weapons.
			if global.traps[trap].trap_type==TRAP_ONCE or
				global.traps[trap].trap_type==TRAP_SWITCH then
			
				if global.traps[trap].player==TRAP_COLLISION_WEAPON then
					handle_trap_collisions(trap, false, 0, false)				
				else
					handle_trap_collisions(trap, true, 0, false)
					-- Set special case that the trap has been activated by
					-- a weapon.
					global.traps[trap].player = TRAP_COLLISION_WEAPON
				end
			
			end

			break
		end
	end	

end	

-- Returns true if the weapon has collided and needs to be removed.
function handle_weapon_collision(weapon_rect, weapon_table)

	local remove = false
	local wrapped = false

	-- See if weapon has wrapped around
	wrapped, weapon_rect = handle_wrap_around(weapon_rect, weapon_table.direction)
		
	if wrapped then
		-- Set the weapon rectangle to the wrapped rectangle.
		weapon_table.left = weapon_rect.left
		weapon_table.top = weapon_rect.top
		weapon_table.right = weapon_rect.right
		weapon_table.bottom = weapon_rect.bottom
		
		--dprintf("weapon_rect after %d %d %d %d", weapon_rect.left, weapon_rect.top, weapon_rect.right, weapon_rect.bottom)

	else
		-- Sanity check new weapon co-ordinates.
		if weapon_rect.left < 0 then
			weapon_rect.left = global.weapons[index].left
		end
		
		if weapon_rect.top < 0 then
			weapon_rect.top = global.weapons[index].top
		end
		
		if weapon_rect.right >= MAX_TILE_WIDTH then
			weapon_rect.right = MAX_TILE_WIDTH - 1
		end		
		
		if weapon_rect.bottom >= MAX_TILE_HEIGHT then
			weapon_rect.bottom = MAX_TILE_HEIGHT - 1
		end
	end	

	local hot_x = weapon_direction_offsets[weapon_table.direction].hot_x
	local hot_y = weapon_direction_offsets[weapon_table.direction].hot_y
	
	local map_x = math.floor((weapon_rect.left + hot_x) / TILE_WIDTH) + 1
	local map_y = math.floor((weapon_rect.top + hot_y) / TILE_HEIGHT) + 1
	
	local map_value = global.map[map_y][map_x]
	
	--dprintf("my: %d mx: %d v: %d", map_y, map_x, map_value)
	
	-- map_value should now contain the value we need to collison check.		
	if map_value>=LEVEL_SPECIAL_RANGE then
		
		-- Special tile collisions.
		local special = map_value - LEVEL_SPECIAL_RANGE
		
		if special==GRAPHIC_SPECIAL_GENERATOR_ON or
			special==GRAPHIC_SPECIAL_GENERATOR_OFF or
			special==GRAPHIC_SPECIAL_GENERATOR_DAMAGED then

			hit_generator(map_x, map_y, weapon_table.player)

			remove = true

		elseif special==GRAPHIC_SPECIAL_GROUND_FORCEFIELD then
			-- Hit a forcefield.				
			if weapon_table.player~=PLAYER_ELF then
				-- Only the Elf's weapon can pass through forcefields.
				remove = true	
			end

		elseif special==GRAPHIC_SPECIAL_GROUND_TRAP then
			-- Hit a trap.		
			if weapon_table.player==PLAYER_ELF then
				-- Only the Elf's weapon can activate traps.
				hit_trap(map_x, map_y)
				-- Let the arrow carry on.
				--remove = true	
			end
		end				
		
	elseif map_value>=LEVEL_EXIT_RANGE then
		
		-- Exit tile collisions.
		local exit = map_value - LEVEL_EXIT_RANGE

		if exit==GRAPHIC_EXIT then

			dprintf("Hit exit.")
			remove = true
			
		elseif exit>=GRAPHIC_EXIT_1 and exit<=GRAPHIC_EXIT_8 then
			
			--dprintf("Entering teleport exit.")
			
			-- Work out the exit number.
			local exit_type = exit - GRAPHIC_EXIT
			
			for area = 1, #global.areas do
				if exit_type==global.areas[area].id then
					
					-- Move the weapon to the teleport end point and
					-- mirror the direction.
					
--					dprintf("Weapon teleported, player %d", weapon_table.player)
					
					local mirror_direction = weapon_direction_offsets[weapon_table.direction].mirror
					weapon_table.direction = mirror_direction
					weapon_table.quad = global.players[weapon_table.player].first_weapon_quad + mirror_direction
					weapon_table.x_offset = weapon_direction_offsets[mirror_direction].x / WEAPON_SCALE_X
					weapon_table.y_offset = weapon_direction_offsets[mirror_direction].y / WEAPON_SCALE_Y						
					weapon_table.left = global.areas[area].left
					weapon_table.top = global.areas[area].top
					weapon_table.right = weapon_table.left + (TILE_WIDTH - 1)
					weapon_table.bottom = weapon_table.top + (TILE_HEIGHT - 1)
					
					break
				end
			end
		end
	else

		-- Wall tile collision.
		local wall = map_value - LEVEL_WALL_RANGE
		
		if wall==GRAPHIC_WALL_DESTRUCTABLE_1 then
			global.map[map_y][map_x] = GRAPHIC_WALL_DESTRUCTABLE_2 + LEVEL_WALL_RANGE
		elseif wall==GRAPHIC_WALL_DESTRUCTABLE_2 then
			global.map[map_y][map_x] = GRAPHIC_WALL_DESTRUCTABLE_3 + LEVEL_WALL_RANGE
		elseif wall==GRAPHIC_WALL_DESTRUCTABLE_3 then
			global.map[map_y][map_x] = LEVEL_SPECIAL_RANGE + GRAPHIC_SPECIAL_GROUND
		elseif wall_2==GRAPHIC_WALL_DESTRUCTABLE_3 then
			global.map[map_y][map_x] = LEVEL_SPECIAL_RANGE + GRAPHIC_SPECIAL_GROUND
		end
		
		remove = true
	end	
	
	if not remove then
		-- Handle weapon collisions with enemies.
		for enemy = 1, #global.enemies do
		
			if rect_intersect_rect(weapon_rect, global.enemies[enemy]) then
								
				-- You can't shoot Death!
				if global.enemies[enemy].enemy_type~=ENEMY_DEATH then
					-- Killed enemy. Remove it from the table.
					table.remove(global.enemies, enemy)
				end
				
				remove = true
				-- Weapon can only kill one enemy.
				break
			end
		end
	end

	return remove
end
	
function update_weapons()

	local weapons = #global.weapons
	for index = 1, weapons do

		local remove = false
--		local wrapped = false
		local weapon_rect = {}
		local x_offset = global.weapons[index].x_offset
		local y_offset = global.weapons[index].y_offset

		weapon_rect.left = global.weapons[index].left + x_offset
		weapon_rect.top = global.weapons[index].top + y_offset
		weapon_rect.right = global.weapons[index].right + x_offset
		weapon_rect.bottom = global.weapons[index].bottom + y_offset

		remove = handle_weapon_collision(weapon_rect, global.weapons[index])
		
		if not remove then
			-- Update the weapon position.
			global.weapons[index].left = global.weapons[index].left + global.weapons[index].x_offset
			global.weapons[index].top = global.weapons[index].top + global.weapons[index].y_offset
			global.weapons[index].right = (global.weapons[index].left + TILE_WIDTH) - 1
			global.weapons[index].bottom = (global.weapons[index].top + TILE_HEIGHT) - 1
		end
			
		global.weapons[index].remove = remove
	end
	
	-- Second pass. Any weapons marked as remove will be deleted.
	for index = weapons, 1, -1 do	
		if global.weapons[index].remove then
			table.remove(global.weapons, index)
		end
	end
	
end
	
function handle_weapon(direction)
		
	local fired = false
	
	if global.players[global.current_player].firing_time then
		-- Check enough time has passed between this new shot and the previous one.
		-- Otherwise we print the two shots on top of each other.
		if global.current_time - global.players[global.current_player].firing_time <
			WEAPON_TIME_INTERVAL then
			return
		end

	end
	
	-- Store the time we fired.
	global.players[global.current_player].firing_time = global.current_time
	
	local weapon_x_offset = 0
	local weapon_y_offset = 0
	
	-- Position weapon a tile away from our current diection.	
	local new_x = global.players[global.current_player].left
	local new_y = global.players[global.current_player].top

	local weapon = {}
	weapon.left = new_x + weapon_direction_offsets[direction].x
	weapon.top = new_y + weapon_direction_offsets[direction].y
	-- Shrink bounding rectangle by 1 pixel - otherwise we go into the next tile if we're
	-- on a co-ordinate offset +/- 2 pixels away from the next tile.	
	weapon.right = weapon.left + (TILE_WIDTH - 1)
	weapon.bottom = weapon.top + (TILE_HEIGHT - 1)
	-- !!!
	weapon.x_offset = weapon_direction_offsets[direction].x / WEAPON_SCALE_X
	weapon.y_offset = weapon_direction_offsets[direction].y / WEAPON_SCALE_Y
	weapon.direction = direction
	weapon.quad = global.players[global.current_player].first_weapon_quad + direction
	weapon.player = global.current_player
	
	--dprintf("Weapon: %d %d d %d q %d %d %d", weapon.left, weapon.top, weapon.direction, weapon.quad, weapon.x_offset, weapon.y_offset)
	
	-- Check to see if where we're initially placing the weapon doesn't create an initial collision.
	-- e.g. we're standing next to a wall and press fire.
	local remove = handle_weapon_collision(weapon, weapon)
	if not remove then
		-- Good to go, so add the weapon.
		local weapon_count = #global.weapons
		global.weapons[weapon_count + 1] = weapon
		--dprintf("Weapon %d added Fired by player %d direction %d", weapon_count + 1, weapon.player, direction)
		fired = true
	end

	return fired
end

